/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

/* Solution of a linear system with STRUMPACK, C interface.
 *
 * This code can work in real and complex arithmetic, single
 * and double precision. However, contratry to the generic
 * C++, this involves changing not only the type of the input
 * matrices and vectors, but also the name of the functions
 * to be called, so we use some macros.
 */
#include "StrumpackDensePackage.h"

#define ARITH_D 0
#define ARITH_S 1
#define ARITH_Z 2
#define ARITH_C 3

/* Choose arithmetic here */
#define ARITH ARITH_D

#if ARITH==ARITH_D
#define myscalar double
#define myreal double
#elif ARITH==ARITH_S
#define myscalar float
#define myreal float
#elif ARITH==ARITH_Z
#define myscalar dcomplex
#define myreal double
#else
#define myscalar scomplex
#define myreal float
#endif

/* The routine names in the C interface depend on the
 * arithmetic so we define some macros to be able to
 * change arithmetic easily. We use concatenations
 * for the nane of the solver structure/object
 * (StrumpackDensePackage_C_[arith]) and for
 * the name of functions (e.g., SDP_C_[arith]_init).
 * Note that for the "func" macro we need an indirection
 * (funceval) because of the way the preprocessor works.
 */
#define strucaux(x) StrumpackDensePackage_C_ ## x
#define struc(x) strucaux(x)
#define funcaux(x,y) SDP_C_ ## y ## _ ## x
#define funceval(x,y) funcaux(x,y)
#define func(x) funceval(x,myscalar)

int main (int argc, char *argv[]) {
  myscalar *A=NULL, *X=NULL, *B=NULL;
  int descA[BLACSCTXTSIZE], descXB[BLACSCTXTSIZE];
  int n;
  int nrhs;
  int nb;
  int locr, locc;
  int i, j, ii, jj;
  int ierr;
  int dummy;
  int myid, np;
  int myrow, mycol, nprow, npcol;
  int ctxt;
  struc(myscalar) SDP_C;

  n=1024; /* Size of the problem */
  nrhs=3; /* Number of RHS */
  nb=16;  /* Blocksize for the 2D block-cyclic distribution */

  /* Initialize MPI */
  if((ierr=MPI_Init(&argc,&argv)))
    return 1;
  myid=-1;
  if((ierr=MPI_Comm_rank(MPI_COMM_WORLD,&myid)))
    return 1;
  np=-1;
  if((ierr=MPI_Comm_size(MPI_COMM_WORLD,&np)))
    return 1;

  /* Initialize the BLACS grid */
  nprow=floor(sqrt((float)np));
  npcol=np/nprow;
  blacs_get_(&IZERO,&IZERO,&ctxt);
  blacs_gridinit_(&ctxt,"R",&nprow,&npcol);
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

  /* A is a dense n x n distributed Toeplitz matrix */
  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    A=malloc(locr*locc*sizeof(myscalar));
    dummy=locr>1?locr:1;
    descinit_(descA,&n,&n,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

    for(i=1;i<=locr;i++)
      for(j=1;j<=locc;j++) {
        ii=indxl2g_(&i,&nb,&myrow,&IZERO,&nprow);
        jj=indxl2g_(&j,&nb,&mycol,&IZERO,&npcol);
        // Toeplitz matrix from Quantum Chemistry.
        myreal pi=3.1416, d=0.1;
#if (ARITH==ARITH_Z) || (ARITH==ARITH_C)
        A[locr*(j-1)+(i-1)].r=ii==jj?pow(pi,2.0)/6.0/pow(d,2.0):pow(-1.0,ii-jj)/pow(ii-jj,2.0)/pow(d,2.0);
        A[locr*(j-1)+(i-1)].i=0;
#else
        A[locr*(j-1)+(i-1)]=ii==jj?pow(pi,2.0)/6.0/pow(d,2.0):pow(-1.0,ii-jj)/pow(ii-jj,2.0)/pow(d,2.0);
#endif
      }
  } else {
    descset_(descA,&n,&n,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Initialize the solver and set parameters */
  func(init)(&SDP_C,MPI_COMM_WORLD);
  SDP_C.use_HSS=1;
  SDP_C.levels_HSS=4;
  SDP_C.min_rand_HSS=10;
  SDP_C.lim_rand_HSS=5;
  SDP_C.inc_rand_HSS=10;
  SDP_C.max_rand_HSS=100;
  SDP_C.tol_HSS=1e-12;
  SDP_C.steps_IR=10;
  SDP_C.tol_IR=1e-10;

  /* Compression */
  func(compress_A)(&SDP_C,A,descA);

  /* Accuracy checking */
  func(check_compression)(&SDP_C,A,descA);

  /* Factorization */
  func(factor)(&SDP_C,A,descA);

  /* Set the RHS (random vector) and the solution space */
  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
    B=malloc(locr*locc*sizeof(myscalar));
    dummy=locr>1?locr:1;
    descinit_(descXB,&n,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);
    for(i=0;i<locr*locc;i++) {
#if (ARITH==ARITH_Z) || (ARITH==ARITH_C)
      B[i].r=((myreal)rand())/((myreal)RAND_MAX);
      B[i].i=0;
#else
      B[i]=((myscalar)rand())/((myscalar)RAND_MAX);
#endif
    }
    X=malloc(locr*locc*sizeof(myscalar));
  } else {
    descset_(descXB,&n,&nrhs,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Triangular solution */
  func(solve)(&SDP_C,X,descXB,B,descXB);

  /* Accuracy checking */
  func(check_solution)(&SDP_C,A,descA,X,descXB,B,descXB);

  /* Iterative refinement */
  func(refine)(&SDP_C,A,descA,X,descXB,B,descXB);

  /* Statistics */
  func(print_statistics)(&SDP_C);

  /* Clean-up */
  free(A);
  free(B);
  free(X);

  /* The end */
  func(destroy)(&SDP_C);
  MPI_Finalize();
  return 0;

}
