/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

/* Using StrumpackDensePackage with a subcommunicator */
#include<StrumpackDensePackage.hpp>

#define myscalar scomplex
#define myreal float

int main (int argc, char *argv[]) {
  myscalar *A=NULL;
  int descA[BLACSCTXTSIZE];
  int n;
  int nb;
  int locr, locc;
  int i, j, ii, jj;
  int ierr;
  int dummy;
  int myid, np;
  int myrow, mycol, nprow, npcol;
  int ctxt;
  int excl[3]={0,4,5};
  MPI_Comm comm;
  MPI_Group base_group, new_group;

  n=256; /* Size of the problem */
  nb=16;  /* Blocksize for the 2D block-cyclic distribution */

  /* Initialize MPI */
  if((ierr=MPI_Init(&argc,&argv)))
    return 1;
  myid=-1;
  if((ierr=MPI_Comm_rank(MPI_COMM_WORLD,&myid)))
    return 1;
  np=-1;
  if((ierr=MPI_Comm_size(MPI_COMM_WORLD,&np)))
    return 1;

  if(np<6) {
    if(!myid) std::cout << "This example requires at least 6 MPI tasks. Exiting." << std::endl;
    return 1;
  }

  ierr=MPI_Comm_group(MPI_COMM_WORLD,&base_group);
  ierr=MPI_Group_excl(base_group,3,excl,&new_group);
  ierr=MPI_Comm_create(MPI_COMM_WORLD,new_group,&comm);

  if(comm!=MPI_COMM_NULL) {
    /* All the processes except 2,4,5 work here */
    MPI_Comm_rank(comm,&myid);
    MPI_Comm_size(comm,&np);

    /* Initialize the BLACS grid */
    nprow=floor(sqrt((float)np));
    npcol=np/nprow;
    ctxt=Csys2blacs_handle(comm);
    Cblacs_gridinit(&ctxt,"R",nprow,npcol);
    Cblacs_gridinfo(ctxt,&nprow,&npcol,&myrow,&mycol);

    /* A is a dense n x n distributed Toeplitz matrix */
    if(myid<nprow*npcol) {
      locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
      locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
      A=new myscalar[locr*locc];
      dummy=std::max(1,locr);
      descinit_(descA,&n,&n,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

      for(i=1;i<=locr;i++)
        for(j=1;j<=locc;j++) {
          ii=indxl2g_(&i,&nb,&myrow,&IZERO,&nprow);
          jj=indxl2g_(&j,&nb,&mycol,&IZERO,&npcol);
          // Toeplitz matrix from Quantum Chemistry.
          myreal pi=3.1416, d=0.1;
          A[locr*(j-1)+(i-1)]=ii==jj?std::pow(pi,2)/6.0/std::pow(d,2):std::pow(-1.0,ii-jj)/std::pow((myreal)ii-jj,2)/std::pow(d,2);
          // Make matrix unsymmetric to check that A^T*X works.
          if(ii>jj)A[locr*(j-1)+(i-1)]*=-1;
        }

    } else {
      descset_(descA,&n,&n,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
    }

    /* Initialize the solver and set parameters */
    StrumpackDensePackage<myscalar,myreal> sdp(comm);
    sdp.use_HSS=true;
    sdp.min_rand_HSS=32;
    sdp.tol_HSS=1e-6;

    /* Compression */
    sdp.compress(A,descA);

    /* Accuracy checking */
    sdp.check_compression(A,descA);

    /* Clean-up */
    delete[] A;
  } else {
    /* Processes 0,4,5 do nothing */
    std::cout << " I am process " << myid << " and I am not working" << std::endl;
  }

  /* The end */
  MPI_Finalize();
  return 0;

}
