!
! STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
! the University of California, through Lawrence Berkeley National Laboratory
! (subject to receipt of any required approvals from the U.S. Dept. of Energy).
! All rights reserved.
!
! If you have questions about your rights to use or distribute this software,
! please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
!
! NOTICE. This software is owned by the U.S. Department of Energy. As such, the
! U.S. Government has been granted for itself and others acting on its behalf a
! paid-up, nonexclusive, irrevocable, worldwide license in the Software to
! reproduce, prepare derivative works, and perform publicly and display publicly.
! Beginning five (5) years after the date permission to assert copyright is
! obtained from the U.S. Department of Energy, and subject to any subsequent five
! (5) year renewals, the U.S. Government is granted for itself and others acting
! on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
! Software to reproduce, prepare derivative works, distribute copies to the
! public, perform publicly and display publicly, and to permit others to do so.
!
! Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
!             (Lawrence Berkeley National Lab, Computational Research Division).
!

! Solution of a linear system with STRUMPACK, Fortran interface.

! Solution of a linear system with STRUMPACK, C interface.
!
! This code can work in real and complex arithmetic, single
! and dcomplex precision. However, contratry to the generic
! C++, this involves changing not only the type of the input
! matrices and vectors, but also the name of the functions
! to be called, so we use some macros.

program F90_example

  use StrumpackDensePackage
  use MPI

  implicit none

  type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
  integer, target :: descA(BLACSCTXTSIZE), descXB(BLACSCTXTSIZE)
  integer :: n, nrhs
  integer :: nb
  integer :: locr, locc
  integer :: i, j, ii, jj
  integer :: ierr, dummy
  integer :: myid, np
  integer :: myrow, mycol, nprow, npcol
  integer :: ctxt
  complex(8), allocatable, target :: A(:), B(:), X(:)
  real(8) :: pi=3.1416D0, d=0.1D0, err
  type(C_PTR) :: CdescA, CdescXB, CA, CB, CX

  n=1024 ! Size of the problem
  nrhs=3 ! Number of RHS
  nb=16  ! Blocksize of the 2D block-cyclic distribution

  ! Initialize MPI
  call MPI_Init(ierr)
  call MPI_Comm_rank(MPI_COMM_WORLD,myid,ierr)
  call MPI_Comm_size(MPI_COMM_WORLD,np,ierr)

  ! Initialize the BLACS grid
  nprow=floor(sqrt(real(np)))
  npcol=np/nprow
  call blacs_get(IZERO,IZERO,ctxt)
  call blacs_gridinit(ctxt,'R',nprow,npcol)
  call blacs_gridinfo(ctxt,nprow,npcol,myrow,mycol)

  ! A is a dense n x n distributed Toeplitz matrix
  if(myid<nprow*npcol) then
    locr=numroc(n,nb,myrow,IZERO,nprow)
    locc=numroc(n,nb,mycol,IZERO,npcol)
    allocate(A(locr*locc))
    A=0
    dummy=max(1,locr)
    call descinit(descA,n,n,nb,nb,IZERO,IZERO,ctxt,dummy,ierr)

    do i=1,locr
      do j=1,locc
        ii=indxl2g(i,nb,myrow,IZERO,nprow)
        jj=indxl2g(j,nb,mycol,IZERO,npcol)
        ! Toeplitz matrix from Quantum Chemistry.
        if(ii==jj) then
          A(locr*(j-1)+i)=pi**2/6.0/d**2
        else
          A(locr*(j-1)+i)=(-1D0)**(ii-jj)/(ii-jj)**2D0/d**2D0
        end if
      end do
    end do
  else
    call descset(descA,n,n,nb,nb,IZERO,IZERO,INONE,IONE)
  end if
  CA=C_LOC(A)
  CdescA=C_LOC(descA)

  ! Initialize the solver and set parameters */
  call SDP_F90_dcomplex_init(SDP_F90,MPI_COMM_WORLD)
  SDP_F90%use_HSS=1
  SDP_F90%levels_HSS=4
  SDP_F90%min_rand_HSS=10
  SDP_F90%lim_rand_HSS=5
  SDP_F90%inc_rand_HSS=10
  SDP_F90%max_rand_HSS=100
  SDP_F90%tol_HSS=1D-12
  SDP_F90%steps_IR=10
  SDP_F90%tol_IR=1D-10

  ! Compression
  call SDP_F90_dcomplex_compress_A(SDP_F90,CA,CdescA)

  ! Accuracy checking
  err=SDP_F90_dcomplex_check_compression(SDP_F90,CA,CdescA)

  ! Factorization
  call SDP_F90_dcomplex_factor(SDP_F90,CA,CdescA)

  ! Set the RHS (random vector) and the solution space
  if(myid<nprow*npcol) then
    locr=numroc(n,nb,myrow,IZERO,nprow)
    locc=numroc(nrhs,nb,mycol,IZERO,npcol)
    allocate(B(locr*locc))
    dummy=max(1,locr)
    call descinit(descXB,n,nrhs,nb,nb,IZERO,IZERO,ctxt,dummy,ierr)
    B=1D0
    allocate(X(locr*locc))
  else
    call descset(descXB,n,nrhs,nb,nb,IZERO,IZERO,INONE,IONE)
  end if
  CX=C_LOC(X)
  CB=C_LOC(B)
  CdescXB=C_LOC(descXB)

  ! Triangular solution */
  call SDP_F90_dcomplex_solve(SDP_F90,CX,CdescXB,CB,CdescXB)

  ! Accuracy checking */
  err=SDP_F90_dcomplex_check_solution(SDP_F90,CA,CdescA,CX,CdescXB,CB,CdescXB)

  ! Iterative refinement */
  err=SDP_F90_dcomplex_refine(SDP_F90,CA,CdescA,CX,CdescXB,CB,CdescXB)

  ! Statistics */
  call SDP_F90_dcomplex_print_statistics(SDP_F90)

  ! Clean-up
  if(myid<nprow*npcol) then
    deallocate(A)
    deallocate(B)
    deallocate(X)
  end if

  ! The end
  call SDP_F90_dcomplex_destroy(SDP_F90)
  call MPI_Finalize(ierr)

end program F90_example
