/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

/* Solution of a linear system with STRUMPACK in a matrix-free situation.
 * The matrix is A = alpha I + U V^T but it is not defined/allocated explicitly.
 * Instead, we define a matrix-vector product routine (Amult) and an element
 * access routine (Aelem).
 *
 * We define random vectors and samples ourselves (Sr = A Rr, Sc = A^* Rc), and
 * we pass them to the compression routine, along with the element access routine.
 * Note that we maintain two versions of the U and V vectors: a distributed version
 * (2D block-cyclic form) used for multiplications ( X <- alpha X + U (V^T X) ),
 * and a replicated version used in Aelem.
 *
 * An important aspect here is that Aelem does not involve interprocess communication.
 * Therefore, we set the access_is_global parameter to false. This improves performance.
 */
#include "StrumpackDensePackage.hpp"

#define myscalar double
#define myreal double
#define MY_MPI_SCALAR MPI_DOUBLE
#define MY_MPI_REAL MPI_DOUBLE

myscalar ZERO=0;
myscalar  ONE=1;

typedef struct objptrs {
  int n;
  int rk;
  int *descUV;
  myscalar alpha;
  myscalar *U;
  myscalar *V;
  myscalar *Udup;
  myscalar *Vdup;
} objptrs;

void Amult(void *obj, char TransA, myscalar *X, myscalar *B, int *descXB) {
  /* Matrix-vector product:
   *    B = A   X = alpha X + U V^T X
   * or B = A^T X = alpha X + V U^T X
   */
  objptrs *ptr;
  int n, rk;
  int nrhs;
  int ctxt;
  int myrow, mycol, nprow, npcol;
  int nb;
  int locr, locc;
  int idummy, ierr;
  int desctmp[BLACSCTXTSIZE];
  int *descUV;
  myscalar *U=NULL, *V=NULL;
  myscalar *tmp;
  myscalar alpha;

  /* Get info from object handle */
  ptr=(objptrs*)obj;
  n=ptr->n;
  rk=ptr->rk;
  alpha=ptr->alpha;
  descUV=ptr->descUV;
  if(TransA=='N') {
    U=ptr->U;
    V=ptr->V;
  } else {
    /* For transposed product, simply swap U and V */
    U=ptr->V;
    V=ptr->U;
  }

  /* Get info from descriptor */
  nrhs=descXB[BLACSn];
  ctxt=descXB[BLACSctxt];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
  nb=descXB[BLACSnb];

  /* tmp = V^T X */
  locr=numroc_(&rk,&nb,&myrow,&IZERO,&nprow);
  locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
  tmp=new myscalar[locr*locc]();
  idummy=std::max(1,locr);
  descinit_(desctmp,&rk,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
  pgemm('C','N',rk,nrhs,n,ONE,V,IONE,IONE,descUV,X,IONE,IONE,descXB,ZERO,tmp,IONE,IONE,desctmp);

  /* B = U tmp */
  pgemm('N','N',n,nrhs,rk,ONE,U,IONE,IONE,descUV,tmp,IONE,IONE,desctmp,ZERO,B,IONE,IONE,descXB);
  delete[] tmp;

  /* B+= alpha X */
  pgeadd('N',n,nrhs,alpha,X,IONE,IONE,descXB,ONE,B,IONE,IONE,descXB);
}

void Aelem(void *obj, int *I, int *J, myscalar *B, int *descB) {
  /* Element access routine */
  objptrs *ptr;
  int n, rk;
  int nI, nJ;
  int ctxt;
  int myrow, mycol, nprow, npcol;
  int nb;
  int locr, locc;
  int i, j, ii, jj, iii, jjj, k;
  myscalar *U=NULL, *V=NULL;
  myscalar alpha;

  /* Get info from object handle */
  ptr=(objptrs*)obj;
  n=ptr->n;
  rk=ptr->rk;
  alpha=ptr->alpha;
  U=ptr->Udup; /* Use the duplicated version of U and V */
  V=ptr->Vdup;

  /* Get info from descriptor */
  nI=descB[BLACSm];
  nJ=descB[BLACSn];
  ctxt=descB[BLACSctxt];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
  nb=descB[BLACSnb];

  locr=numroc_(&nI,&nb,&myrow,&IZERO,&nprow);
  locc=numroc_(&nJ,&nb,&mycol,&IZERO,&npcol);
  for(i=1;i<=locr;i++) {
    for(j=1;j<=locc;j++) {
      ii=indxl2g_(&i,&nb,&myrow,&IZERO,&nprow);
      jj=indxl2g_(&j,&nb,&mycol,&IZERO,&npcol);
      iii=I[ii-1];
      jjj=J[jj-1];
      /* B_loc(i,j) = B(ii,jj) = A(iii,jjj) = alpha delta_{iii,jjj} + sum_{k=1}^{rk} U(iii,k)*V(jjj,k) */
      B[locr*(j-1)+(i-1)]=ZERO;
      for(k=0;k<rk;k++)
        B[locr*(j-1)+(i-1)]+=U[n*k+iii-1]*V[n*k+jjj-1];
      if(iii==jjj)
        B[locr*(j-1)+(i-1)]+=alpha;
    }
  }
}

int main (int argc, char *argv[]) {
  myscalar *Rr=NULL, *Rc=NULL, *Sr=NULL, *Sc=NULL;
  myscalar *Xtrue=NULL, *X=NULL, *B=NULL;
  myscalar *Udup=NULL, *Vdup=NULL;
  myscalar *U=NULL, *V=NULL;
  myscalar alpha;
  myreal maxtrue, maxdiff;
  int descUV[BLACSCTXTSIZE], descRS[BLACSCTXTSIZE], descXB[BLACSCTXTSIZE], desccent[BLACSCTXTSIZE];
  int n;
  int nrhs;
  int rk, rd;
  int nb;
  int locr, locc;
  int i;
  int ierr;
  int dummy;
  int myid, np;
  int myrow, mycol, nprow, npcol;
  int ctxt, ctxtcent, ctxtglob;
  objptrs ptr;

  n=1024;    /* Size of the problem */
  nrhs=3;    /* Number of RHS */
  nb=16;     /* Blocksize for the 2D block-cyclic distribution */
  rk=17;     /* Rank of U and V */
  rd=32;     /* User-given sample size */
  alpha=10;  /* Defines A */

  /* Initialize MPI */
  if((ierr=MPI_Init(&argc,&argv)))
    return 1;
  myid=-1;
  if((ierr=MPI_Comm_rank(MPI_COMM_WORLD,&myid)))
    return 1;
  np=-1;
  if((ierr=MPI_Comm_size(MPI_COMM_WORLD,&np)))
    return 1;

  /* Initialize a BLACS grid that contains only id 0, for redistribution */
  blacs_get_(&IZERO,&IZERO,&ctxtcent);
  blacs_gridinit_(&ctxtcent,"R",&IONE,&IONE);

  /* We initialize a context with all the processes */
  blacs_get_(&IZERO,&IZERO,&ctxtglob);
  blacs_gridinit_(&ctxtglob,"R",&IONE,&np);

  /* Initialize the BLACS grid */
  nprow=floor(sqrt((float)np));
  npcol=np/nprow;
  blacs_get_(&IZERO,&IZERO,&ctxt);
  blacs_gridinit_(&ctxt,"R",&nprow,&npcol);
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

  /* Generate U and V: we have a 2D block-cyclic version and a duplicated version.
   * Id 0 generates the values in Udup and Vdup, then it broadcasts them
   * to other processes and also redistributes them into a 2D block-cyclic version.
   */
  Udup=new myscalar[n*rk];
  Vdup=new myscalar[n*rk];
  if(!myid) {
    for(i=0;i<n*rk;i++) {
      Udup[i]=static_cast<myscalar>(rand())/(static_cast<myscalar>(RAND_MAX));
      Vdup[i]=static_cast<myscalar>(rand())/(static_cast<myscalar>(RAND_MAX));
    }
    descinit_(desccent,&n,&rk,&nb,&nb,&IZERO,&IZERO,&ctxtcent,&n,&ierr);
  } else {
    descset_(desccent,&n,&rk,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }
  MPI_Bcast((void *)Udup,n*rk,MY_MPI_SCALAR,IZERO,MPI_COMM_WORLD);
  MPI_Bcast((void *)Vdup,n*rk,MY_MPI_SCALAR,IZERO,MPI_COMM_WORLD);

  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&rk,&nb,&mycol,&IZERO,&npcol);
    dummy=std::max(1,locr);
    descinit_(descUV,&n,&rk,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);
    U=new myscalar[locr*locc];
    V=new myscalar[locr*locc];
  } else {
    descset_(descUV,&n,&rk,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }
  pgemr2d(n,rk,Udup,IONE,IONE,desccent,U,IONE,IONE,descUV,ctxtglob);
  pgemr2d(n,rk,Vdup,IONE,IONE,desccent,V,IONE,IONE,descUV,ctxtglob);

  /* Pack n, rk, alpha, U, V, descUV into a single object to be given to STRUMPACK */
  ptr.n=n;
  ptr.rk=rk;
  ptr.alpha=alpha;
  ptr.descUV=descUV;
  ptr.U=U;
  ptr.V=V;
  ptr.Udup=Udup;
  ptr.Vdup=Vdup;

  /* Rr and Rc are random vectors, Sr and Sc are samples
   * of the row and column space of A.
   */
  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&rd,&nb,&mycol,&IZERO,&npcol);
    dummy=std::max(1,locr);
    descinit_(descRS,&n,&rd,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

    /* Different seeds */
    srand(time(NULL)+myid);

    Rr=new myscalar[locr*locc];
    Rc=new myscalar[locr*locc];
    for(i=0;i<locr*locc;i++) {
      Rr[i]=static_cast<myscalar>(rand())/(static_cast<myscalar>(RAND_MAX));
      Rc[i]=static_cast<myscalar>(rand())/(static_cast<myscalar>(RAND_MAX));
    }

    Sr=new myscalar[locr*locc]();
    Sc=new myscalar[locr*locc]();
    Amult((void *)&ptr,'N',Rr,Sr,descRS);
    Amult((void *)&ptr,'C',Rc,Sc,descRS);

  } else {
    descset_(descRS,&n,&rd,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Initialize the solver and set parameters */
  StrumpackDensePackage<myscalar,myreal> sdp(MPI_COMM_WORLD);
  sdp.use_HSS=true;
  sdp.block_HSS=rk;
  sdp.min_rand_HSS=rd;
  sdp.lim_rand_HSS=0;
  sdp.tol_HSS=1e-14;
  sdp.access_is_global=false;
  sdp.obj=&ptr;

  /* Compression */
  sdp.compress(Rr,Rc,Sr,Sc,descRS,Aelem);

  /* Factorization */
  sdp.factor(NULL,NULL);

  /* Set the RHS (random vector) and the solution space */
  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
    dummy=std::max(1,locr);
    descinit_(descXB,&n,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

    Xtrue=new myscalar[locr*locc];
    for(i=0;i<locr*locc;i++)
      Xtrue[i]=static_cast<myscalar>(rand())/(static_cast<myscalar>(RAND_MAX));

    B=new myscalar[locr*locc];
    Amult((void *)&ptr,'N',Xtrue,B,descXB);

    X=new myscalar[locr*locc];
  } else {
    descset_(descXB,&n,&nrhs,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Triangular solution */
  sdp.solve(X,descXB,B,descXB);

  /* Forward error check: compute max(|X-Xtrue|)/max(|Xtrue|) */
  maxtrue=0;
  maxdiff=0;
  if(myid<nprow*npcol) {
    for(i=0;i<locr*locc;i++) {
      maxdiff=std::max(maxdiff,std::abs(X[i]-Xtrue[i]));
      maxtrue=std::max(maxtrue,std::abs(Xtrue[i]));
    }
  }
  MPI_Allreduce(MPI_IN_PLACE,(void*)&maxtrue,IONE,MY_MPI_REAL,MPI_MAX,MPI_COMM_WORLD);
  MPI_Allreduce(MPI_IN_PLACE,(void*)&maxdiff,IONE,MY_MPI_REAL,MPI_MAX,MPI_COMM_WORLD);
  if(!myid)
    std::cout << "Forward error = " << maxdiff/maxtrue << std::endl;

  /* Clean-up */
  delete[] U;
  delete[] V;
  delete[] Udup;
  delete[] Vdup;
  delete[] Rr;
  delete[] Rc;
  delete[] Sr;
  delete[] Sc;
  delete[] B;
  delete[] X;
  delete[] Xtrue;

  /* The end */
  MPI_Finalize();
  return 0;

}
