/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

/* Simplistic power method with STRUMPACK */
#include "StrumpackDensePackage.hpp"

#define myscalar dcomplex
#define myreal double

myscalar ZERO=0;
myscalar  ONE=1;

int main (int argc, char *argv[]) {
  myscalar *A=NULL, *Bprev=NULL, *B=NULL;
  myscalar lambda, lambdaprev;
  myreal tol, err, nrm;
  int descA[BLACSCTXTSIZE], descB[BLACSCTXTSIZE], desc[BLACSCTXTSIZE];
  int n;
  int nb;
  int locr, locc;
  int i, j, ii, jj;
  int ierr;
  int dummy;
  int myid, np;
  int myrow, mycol, nprow, npcol;
  int ctxt;
  int max_steps, steps, go;
  double tstart, tend;

  n=1024; /* Size of the problem */
  nb=16;  /* Blocksize for the 2D block-cyclic distribution */

  /* Initialize MPI */
  if((ierr=MPI_Init(&argc,&argv)))
    return 1;
  myid=-1;
  if((ierr=MPI_Comm_rank(MPI_COMM_WORLD,&myid)))
    return 1;
  np=-1;
  if((ierr=MPI_Comm_size(MPI_COMM_WORLD,&np)))
    return 1;

  /* Initialize the BLACS grid */
  nprow=floor(sqrt((float)np));
  npcol=np/nprow;
  blacs_get_(&IZERO,&IZERO,&ctxt);
  blacs_gridinit_(&ctxt,"R",&nprow,&npcol);
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

  /* A is a dense n x n distributed Toeplitz matrix */
  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    A=new myscalar[locr*locc];
    dummy=std::max(1,locr);
    descinit_(descA,&n,&n,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

    for(i=1;i<=locr;i++)
      for(j=1;j<=locc;j++) {
        ii=indxl2g_(&i,&nb,&myrow,&IZERO,&nprow);
        jj=indxl2g_(&j,&nb,&mycol,&IZERO,&npcol);
        // Simplistic Toeplitz matrix; rank should be 2.
        //A[locr*(j-1)+(i-1)]=ii==jj?n*n:ii-jj;
        // Toeplitz matrix from Quantum Chemistry.
        myreal pi=3.1416, d=0.1;
        A[locr*(j-1)+(i-1)]=ii==jj?std::pow(pi,2)/6.0/std::pow(d,2):std::pow(-1.0,ii-jj)/std::pow((myreal)ii-jj,2)/std::pow(d,2);
      }

  } else {
    descset_(descA,&n,&n,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Initialize the solver and set parameters */
  StrumpackDensePackage<myscalar,myreal> sdp(MPI_COMM_WORLD);
  sdp.use_HSS=true;
  sdp.min_rand_HSS=64;
  sdp.levels_HSS=4;
  sdp.tol_HSS=1e-7;

  /* Compression */
  sdp.compress(A,descA);

  /* Accuracy checking */
  sdp.check_compression(A,descA);

  /* Set the starting vector (random vector) */
  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&IONE,&nb,&mycol,&IZERO,&npcol);
    B=new myscalar[locr*locc];
    dummy=std::max(1,locr);
    descinit_(descB,&n,&IONE,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);
    descinit_(desc,&IONE,&IONE,&nb,&nb,&IZERO,&IZERO,&ctxt,&IONE,&ierr);
    for(i=0;i<locr*locc;i++)
      B[i]=1.0;
    Bprev=new myscalar[locr*locc];
  } else {
    descset_(descB,&n,&IONE,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
    descset_(desc,&IONE,&IONE,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  max_steps=500;
  tol=1e-5;
  go=1;
  steps=0;
  sdp.verbose=false;
  tstart=MPI_Wtime();
  if(!myid) std::cout << "Computing the largest eigenvalue... " << std::endl;
  while(go && steps<max_steps) {
    if(myid<nprow*npcol) {
      lacpy('N',locr,locc,B,locr,Bprev,locr);
      lambdaprev=lambda;
    }
    sdp.product('N',ONE,A,descA,Bprev,descB,ZERO,B,descB);
    if(myid<nprow*npcol) {
      pgemm('C','N',IONE,IONE,n,ONE,Bprev,IONE,IONE,descB,B,IONE,IONE,descB,ZERO,&lambda,IONE,IONE,desc);
      nrm=pnrm2<myscalar,myreal>(n,Bprev,IONE,IONE,descB,IONE);
      lambda=lambda/nrm/nrm;
      nrm=pnrm2<myscalar,myreal>(n,B,IONE,IONE,descB,IONE);
      pscal(n,ONE/nrm,B,IONE,IONE,descB,IONE);
      err=std::abs(lambda-lambdaprev)/std::abs(lambda);
      if(err<tol)
        go=0;
    }
    MPI_Bcast((void *) &go,IONE,MPI_INTEGER,IZERO,MPI_COMM_WORLD);
    steps++;
    if(!myid && steps%25==0) std::cout << "Step: " << steps << "; lambda = " << lambda << "; err = " << err << std::endl;
  }
  tend=MPI_Wtime();
  if(!myid && steps%25!=0) std::cout << "Step: " << steps << "; lambda = " << lambda << "; err = " << err << std::endl;
  if(!myid) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;

  /* Clean-up */
  delete[] A;
  delete[] B;

  /* The end */
  MPI_Finalize();
  return 0;

}
