/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

/* Compressed matrix-vector product with STRUMPACK */
#include "StrumpackDensePackage.hpp"

#define myscalar dcomplex
#define myreal double

int main (int argc, char *argv[]) {
  myscalar *A=NULL, *X=NULL, *B1=NULL, *B2=NULL;
  myreal nrm;
  int descA[BLACSCTXTSIZE], descB[BLACSCTXTSIZE], desc[BLACSCTXTSIZE];
  int n;
  int nb;
  int locr, locc;
  int i, j, ii, jj;
  int ierr;
  int dummy;
  int myid, np;
  int myrow, mycol, nprow, npcol;
  int ctxt;
  myscalar alpha=static_cast<myscalar>(rand()), beta=static_cast<myscalar>(rand());

  n=2048; /* Size of the problem */
  nb=16;  /* Blocksize for the 2D block-cyclic distribution */

  /* Initialize MPI */
  if((ierr=MPI_Init(&argc,&argv)))
    return 1;
  myid=-1;
  if((ierr=MPI_Comm_rank(MPI_COMM_WORLD,&myid)))
    return 1;
  np=-1;
  if((ierr=MPI_Comm_size(MPI_COMM_WORLD,&np)))
    return 1;

  /* Initialize the BLACS grid */
  nprow=floor(sqrt((float)np));
  npcol=np/nprow;
  blacs_get_(&IZERO,&IZERO,&ctxt);
  blacs_gridinit_(&ctxt,"R",&nprow,&npcol);
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

  /* A is a dense n x n distributed Toeplitz matrix */
  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    A=new myscalar[locr*locc];
    dummy=std::max(1,locr);
    descinit_(descA,&n,&n,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

    for(i=1;i<=locr;i++)
      for(j=1;j<=locc;j++) {
        ii=indxl2g_(&i,&nb,&myrow,&IZERO,&nprow);
        jj=indxl2g_(&j,&nb,&mycol,&IZERO,&npcol);
        // Toeplitz matrix from Quantum Chemistry.
        myreal pi=3.1416, d=0.1;
        A[locr*(j-1)+(i-1)]=ii==jj?std::pow(pi,2)/6.0/std::pow(d,2):std::pow(-1.0,ii-jj)/std::pow((myreal)ii-jj,2)/std::pow(d,2);
        // Make matrix unsymmetric to check that A^T*X works.
        if(ii>jj)A[locr*(j-1)+(i-1)]*=-1;
      }

  } else {
    descset_(descA,&n,&n,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Initialize the solver and set parameters */
  StrumpackDensePackage<myscalar,myreal> sdp(MPI_COMM_WORLD);
  sdp.use_HSS=true;
  sdp.min_rand_HSS=64;
  sdp.block_HSS=128;
  sdp.tol_HSS=1e-6;

  /* Compression */
  sdp.compress(A,descA);

  /* Accuracy checking */
  sdp.check_compression(A,descA);

  /* X random vector, B1=B2 random vectors */
  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&IONE,&nb,&mycol,&IZERO,&npcol);
    X=new myscalar[locr*locc];
    dummy=std::max(1,locr);
    descinit_(descB,&n,&IONE,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);
    descinit_(desc,&IONE,&IONE,&nb,&nb,&IZERO,&IZERO,&ctxt,&IONE,&ierr);
    for(i=0;i<locr*locc;i++)
      X[i]=static_cast<myscalar>(rand());
    B1=new myscalar[locr*locc];
    for(i=0;i<locr*locc;i++)
      B1[i]=static_cast<myscalar>(rand());
    B2=new myscalar[locr*locc];
    for(i=0;i<locr*locc;i++)
      B2[i]=B1[i];
  } else {
    descset_(descB,&n,&IONE,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Compute alpha*A*X+beta*B */
  sdp.use_HSS=false;
  sdp.product('N',alpha,A,descA,X,descB,beta,B1,descB);

  /* Compute alpha*HSS(A)*X+beta*B */
  sdp.use_HSS=true;
  sdp.product('N',alpha,A,descA,X,descB,beta,B2,descB);

  if(myid<nprow*npcol) {
    nrm=pnrm2<myscalar,myreal>(n,B1,IONE,IONE,descB,IONE);
    for(i=0;i<locr*locc;i++)
      B2[i]=B2[i]-B1[i];
    nrm=pnrm2<myscalar,myreal>(n,B2,IONE,IONE,descB,IONE)/nrm;
  }
  if(!myid) std::cout << "err=" << nrm << std::endl << std::endl;

  /* Reset the vectors */
  if(myid<nprow*npcol) {
    for(i=0;i<locr*locc;i++)
      B1[i]=static_cast<myscalar>(rand());
    for(i=0;i<locr*locc;i++)
      B2[i]=B1[i];
  }

  /* Compute alpha*A'*X+beta*B */
  sdp.use_HSS=false;
  sdp.product('C',alpha,A,descA,X,descB,beta,B1,descB);

  /* Compute alpha*HSS(A)'*X+beta*B */
  sdp.use_HSS=true;
  sdp.product('C',alpha,A,descA,X,descB,beta,B2,descB);

  if(myid<nprow*npcol) {
    nrm=pnrm2<myscalar,myreal>(n,B1,IONE,IONE,descB,IONE);
    for(i=0;i<locr*locc;i++)
      B2[i]=B2[i]-B1[i];
    nrm=pnrm2<myscalar,myreal>(n,B2,IONE,IONE,descB,IONE)/nrm;
  }
  if(!myid) std::cout << "err=" << nrm << std::endl << std::endl;

  /* Clean-up */
  delete[] X;
  delete[] B1;
  delete[] B2;

  /* The end */
  MPI_Finalize();
  return 0;

}
