/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

/* Solution of a linear system with STRUMPACK, using a hybrid approach.
 * A = [ A11 A12 ]
 *     [ A21 A22 ]
 * and we use the factorization
 * A = [ P11' L11   ,   0 ] * [ U11 ,   L11^-1 P11 A12 ]
 *     [ A21 U11^-1 ,   I ]   [ 0   ,                S ]
 * with S the Schur complement S = A22 - A21 A11^-1 A12.
 *
 * To solve Ax=b, we solve:
 *   a/ [ P11' L11   ,   0 ] * y = b
 *      [ A21 U11^-1 ,   I ]
 *
 *   b/ [ U11 ,   L11^-1 P11 A12 ] * x = y
 *      [ 0   ,                S ]
 *
 * We proceed as follows:
 *   1/ Partial factorization:
 *      1.1/ P11 A11 = L11 U11; we store F11 <-L11+U11-I, and P11 apart.
 *      1.2/ F12 <- L11^-1 P11 A12
 *      1.3/ F21 <- A21 U11^-1
 *   2/ Computation of a Schur complement:
 *      F22 <- A22 - F21 F12
 *   3/ "Reduction" ("condensation") of the RHS:
 *      1/ y1 = L11^-1 P11 b1 = tril(F11) \ b1
 *      2/ y2 = b2 - (A21 U11^-1) y1 = b2 - F21 y1
 *   4/ Solution of the Schur complement system with an iterative method.
 *      Here we use Conjugate Gradients because the system happens to
 *      be SPD.
 *      x2 = S^-1 y2 = F22 \ y2
 *   5/ "Expansion" phase, solution of the initial problem.
 *      x1 = U11^-1 (y1 - L11^-1 P11 A12 x2) = triu(F11) \ (y1 - F12 x2)
 *
 * Using STRUMPACK, this amounts to:
 *   0/ sdp.compress(...)
 *   1/ sdp.partial_factor(...)
 *   2/ sdp.compute_schur(...)
 *   3/ sdp.reduce_RHS(...)
 *   4/ CG(...). This one is provided by the user but uses sdp.schur_product.
 *   5/ sdp.expand_solve(...)
 *
 */
#include "StrumpackDensePackage.hpp"

#define myscalar dcomplex
#define myreal double
#define MY_MPI_REAL MPI_DOUBLE

myscalar NONE=-1;
myscalar ZERO= 0;
myscalar  ONE= 1;

/* Solution routine for the Schur complement system, provided by the user */
void CG(StrumpackDensePackage<myscalar,myreal> *sdp, myscalar *X, myscalar *B, int *descVec, int n, int nrhs, int niter, myreal threshold);

int main (int argc, char *argv[]) {
  myscalar *A=NULL, *X=NULL, *B=NULL, *Y=NULL, *Xtrue=NULL;
  myscalar elem;
  int descA[BLACSCTXTSIZE], descVec[BLACSCTXTSIZE], descelem[BLACSCTXTSIZE];
  int n;
  int nrhs;
  int nb;
  int locr, locc;
  int i, j, ii, jj;
  int ierr;
  int dummy;
  int myid, np;
  int myrow, mycol, nprow, npcol;
  int ctxt;
  myreal rdummy, res;

  n=1024; /* Size of the problem */
  nrhs=3; /* Number of RHS */
  nb=16;  /* Blocksize for the 2D block-cyclic distribution */

  /* Initialize MPI */
  if((ierr=MPI_Init(&argc,&argv)))
    return 1;
  myid=-1;
  if((ierr=MPI_Comm_rank(MPI_COMM_WORLD,&myid)))
    return 1;
  np=-1;
  if((ierr=MPI_Comm_size(MPI_COMM_WORLD,&np)))
    return 1;

  /* Initialize the BLACS grid */
  nprow=floor(sqrt((float)np));
  npcol=np/nprow;
  blacs_get_(&IZERO,&IZERO,&ctxt);
  blacs_gridinit_(&ctxt,"R",&nprow,&npcol);
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

  /* A is a dense n x n distributed Toeplitz matrix */
  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    A=new myscalar[locr*locc];
    dummy=std::max(1,locr);
    descinit_(descA,&n,&n,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

    myreal pi=3.1416, d=0.1;
    for(i=1;i<=locr;i++) {
      for(j=1;j<=locc;j++) {
        ii=indxl2g_(&i,&nb,&myrow,&IZERO,&nprow);
        jj=indxl2g_(&j,&nb,&mycol,&IZERO,&npcol);
        // Toeplitz matrix from Quantum Chemistry.
        A[locr*(j-1)+(i-1)]=ii==jj?std::pow(pi,2)/6.0/std::pow(d,2) : std::pow(-1.0,ii-jj)/std::pow((myreal)ii-jj,2)/std::pow(d,2);
      }
    }
  } else {
    descset_(descA,&n,&n,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Initializing solution */
  if(myid<nprow*npcol) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
    dummy=std::max(1,locr);
    descinit_(descVec,&n,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);
    Xtrue=new myscalar[locr*locc]();
    for(i=0;i<locr*locc;i++)
      Xtrue[i]=static_cast<myscalar>(rand())/(static_cast<myscalar>(RAND_MAX));
  } else {
    descset_(descVec,&n,&nrhs,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Initializing solution and intermediate vector space */
  if(myid<nprow*npcol) {
    X=new myscalar[locr*locc]();
    Y=new myscalar[locr*locc]();
  }

  /* Initializing RHS as A*Xtrue */
  if(myid<nprow*npcol) {
    B=new myscalar[locr*locc]();
    pgemm('N','N',n,nrhs,n,ONE,A,IONE,IONE,descA,Xtrue,IONE,IONE,descVec,ZERO,B,IONE,IONE,descVec);
  }

  /* Initialize the solver and set parameters */
  StrumpackDensePackage<myscalar,myreal> sdp(MPI_COMM_WORLD);
  sdp.use_HSS=true;
  sdp.levels_HSS=4;
  sdp.min_rand_HSS=64;
  sdp.lim_rand_HSS=0;
  sdp.tol_HSS=1e-12;
  sdp.split_HSS=768; /* Size of A11 */

  /* Compression */
  sdp.compress(A,descA);

  /* Accuracy checking */
  sdp.check_compression(A,descA);

  /* Factorization */
  sdp.partially_factor(A,descA,sdp.split_HSS);

  /* Schur complement update */
  sdp.compute_schur();

  /* Extracting a random element from the Schur complement */
  if(myid<nprow*npcol) {
    descinit_(descelem,&IONE,&IONE,&nb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);
  } else {
    descset_(descelem,&IONE,&IONE,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }
  i=1+rand()%(n-sdp.split_HSS);
  j=1+rand()%(n-sdp.split_HSS);
  MPI_Bcast((void *)&i,IONE,MPI_INTEGER,IZERO,MPI_COMM_WORLD);
  MPI_Bcast((void *)&j,IONE,MPI_INTEGER,IZERO,MPI_COMM_WORLD);
  sdp.extract_schur(&elem,descelem,&i,IONE,&j,IONE);
  if(!myid)
    std::cout << "Element (" << i << "," << j << ") of Schur complement = " << elem << std::endl << std::endl;

  /* Condensation */
  sdp.reduce_RHS(Y,descVec,B,descVec);

  /* Solve the Schur complement system (touches only the bottom part of X)*/
  sdp.verbose=false;
  if(!myid)
    std::cout << "Solving Schur complement system with Conjugate Gradient..." << std::endl << std::endl;
  CG(&sdp,X,Y,descVec,n,nrhs,3000,1e-14);
  sdp.verbose=true;

  /* Expansion (touches only the top part of X) */
  sdp.expand_solution(X,descVec,Y,descVec);

  /* Accuracy checking */
  sdp.check_solution(A,descA,X,descVec,B,descVec);

  /* Forward error */
  if(myid<nprow*npcol) {
    for(i=0;i<locr*locc;i++)
      Y[i]=X[i]-Xtrue[i];
    res=plange('F',n,nrhs,Y,IONE,IONE,descVec,&rdummy);
    res/=plange('F',n,nrhs,Xtrue,IONE,IONE,descVec,&rdummy);
    if(!myid)
      std::cout << "Forward error = " << res << std::endl;
  }

  /* Statistics */
  sdp.print_statistics();

  /* Clean-up */
  delete[] A;
  delete[] B;
  delete[] X;
  delete[] Y;
  delete[] Xtrue;

  /* The end */
  MPI_Finalize();
  return 0;

}

void CG(StrumpackDensePackage<myscalar,myreal> *sdp, myscalar *X, myscalar *B, int *descVec, int n, int nrhs, int niter, myreal threshold) {
  /* Conjugate Gradients. Calls sdp.schur_product for matvecs. Fills only the bottom part of X. */
  int printit=250;
  int IA;
  int ctxt;
  int nprow, npcol;
  int myrow, mycol;
  int rsrc, csrc;
  int nb;
  int locr, locc;
  int i, j;
  int idummy, ierr;
  int neff;
  int it;
  int desc[BLACSCTXTSIZE], descScal[BLACSCTXTSIZE];
  bool ingrid;
  myreal res;
  myscalar *x=NULL, *b=NULL;
  myscalar *r=NULL, *p=NULL, *Ap=NULL;
  myscalar alpha, rrprev, rrnext;

  IA=sdp->split_HSS+1;
  neff=n-IA+1;

  ctxt=descVec[BLACSctxt];
  rsrc=descVec[BLACSrsrc];
  csrc=descVec[BLACScsrc];
  nb=descVec[BLACSmb];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
  ingrid=myrow>=0 && mycol>=0;

  /* X and B have n rows. We create versions with
   * the last neff rows only that we pass to SDP.
   */
  if(ingrid) {
    locr=numroc_(&neff,&nb,&myrow,&rsrc,&nprow);
    locc=numroc_(&IONE,&nb,&mycol,&csrc,&npcol);
    idummy=locr>1?locr:1;
    descinit_(desc,&neff,&IONE,&nb,&nb,&rsrc,&csrc,&ctxt,&idummy,&ierr);
    x=new myscalar[locr*locc]();
    b=new myscalar[locr*locc]();
    r=new myscalar[locr*locc]();
    p=new myscalar[locr*locc]();
    Ap=new myscalar[locr*locc]();
  } else {
    locr=0;
    locc=0;
    descset_(desc,&neff,&IONE,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Descriptor for the scalar containing the result of the dot product */
  if(ingrid)
    descinit_(descScal,&IONE,&IONE,&nb,&nb,&IZERO,&IZERO,&ctxt,&IONE,&ierr);
  else
    descset_(descScal,&IONE,&IONE,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);

  for(j=1;j<=nrhs;j++) {
    /* One RHS at a time */
    if(ingrid) {
      pgeadd('N',neff,IONE,ONE,B,IA,j,descVec,ZERO,b,IONE,IONE,desc);
      pgeadd('N',neff,IONE,ONE,X,IA,j,descVec,ZERO,x,IONE,IONE,desc);
    }

    /* r = b - A x
     * If the Schur was explitly in matrix A, the BLAS call would be
     * pgemm('N','N',neff,IONE,neff,NONE,A,IA,IA,descA,x,IONE,IONE,desc,ONE,r,IONE,IONE,desc);
     *
     */
    if(ingrid)
      placpy('N',neff,IONE,b,IONE,IONE,desc,r,IONE,IONE,desc);
    sdp->schur_product('N',NONE,x,desc,ONE,r,desc);

    /* p = r */
    if(ingrid)
      placpy('N',neff,IONE,r,IONE,IONE,desc,p,IONE,IONE,desc);

    /* "Previous" r' * r */
    rrprev=ZERO;
    if(ingrid)
      pgemm('C','N',IONE,IONE,neff,ONE,r,IONE,IONE,desc,r,IONE,IONE,desc,ZERO,&rrprev,IONE,IONE,descScal);
    MPI_Bcast((void *)&rrprev,IONE,MY_MPI_REAL,IZERO,MPI_COMM_WORLD);

    it=1;
    while(it<=niter) {
      /* Ap = A * p
       * If the Schur was explitly in matrix A, the BLAS call would be
       * pgemm('N','N',neff,IONE,neff,ONE,A,IA,IA,descA,p,IONE,IONE,desc,ZERO,Ap,IONE,IONE,desc);
       *
       */
      sdp->schur_product('N',ONE,p,desc,ZERO,Ap,desc);

      /* alpha = r'*r / (p' * A * p) = rrprev/(p' * Ap) */
      alpha=ZERO;
      if(ingrid)
        pgemm('C','N',IONE,IONE,neff,ONE,p,IONE,IONE,desc,Ap,IONE,IONE,desc,ZERO,&alpha,IONE,IONE,descScal);
      MPI_Bcast((void *)&alpha,IONE,MY_MPI_REAL,IZERO,MPI_COMM_WORLD);
      alpha=rrprev/alpha;

      /* x = x + alpha * p */
      for(i=0;i<locr*locc;i++)
        x[i]+=alpha*p[i];

      /* r = r - alpha * Ap */
      for(i=0;i<locr*locc;i++)
        r[i]-=alpha*Ap[i];

      /* "Next" r' * r */
      rrnext=ZERO;
      if(ingrid)
        pgemm('C','N',IONE,IONE,neff,ONE,r,IONE,IONE,desc,r,IONE,IONE,desc,ZERO,&rrnext,IONE,IONE,descScal);
      MPI_Bcast((void *)&rrnext,IONE,MY_MPI_REAL,IZERO,MPI_COMM_WORLD);

      /* Residual */
      res=sqrt(rrnext.real());
      if(it%printit==0)
        if(!myrow && !mycol)
          std::cout << "RHS " << j << ": iteration " << it << ", ||Ax-b||/||b||=" << res << std::endl;
      if(res<threshold)
        break;

      /* p = r + rrnext/rrprev * p */
      for(i=0;i<locr*locc;i++)
        p[i]=r[i]+rrnext/rrprev*p[i];

      /* "Previous" r' * r */
      rrprev=rrnext;

      it++;
    }

    if(it>niter)
      it=niter;

    if(it%printit)
      if(!myrow && !mycol)
        std::cout << "RHS " << j << ": iteration " << it << ", ||Ax-b||/||b||=" << res << std::endl << std::endl;

    /* Back to n-sized vector */
    if(ingrid)
      pgeadd('N',neff,IONE,ONE,x,IONE,IONE,desc,ZERO,X,IA,j,descVec);

  }

  delete[] b;
  delete[] x;
  delete[] r;
  delete[] p;
  delete[] Ap;
}
