/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSSR_SEQ_COMPRESS_HPP
#define HSSR_SEQ_COMPRESS_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSSR_seq_Compress(T *A, S tol, int dmin, int dmax, int dlim, int dinc, int *deff, int *r,int doRoot, T* Rr, T* Rc, T* Sr, T* Sc, StrumpackRand *rgen, void (*elements)(void *, int *, int *, T *, int *), int offset, bool dosync) {
  /* HSS Compression of a dense matrix with randomized sampling.
   * A is a dense n x n matrix.
   * dmin is the number of random vectors to be used; it is used
   * at the beginning of the compression. If, during compression,
   * the HSS rank is found to be too close to the number of
   * random vectors, the number of random vectors is increased.
   * dlim is how close the rank is allowed to be from the number
   * of random vectors.
   * dmax is the maximum number of random vectors to be used.
   * dinc is the increment (number of additional random vectors
   * when expanded).
   * deff is the effective number of random vectors used.
   * r is the maximum rank found during the compression (output).
   * doRoot sets the special treatment of the root node (see below).
   * Rr, Rc, Sr, Sr are random vectors and samplings of A (see below);
   * they are either provided (non-NULL) or computed by the routine.
   * The matrix A can be provided as:
   *  - An explicit matrix,
   *  - An element access routine and explicit samples Sr and Sc,
   *  - An element access routine and a matrix-vector product.
   * When this sequential compression routine is called within
   * a parallel compression, the element access routine needs
   * to be called with global indices (that correspond to the
   * global compression), hence the "offset" parameter. Also,
   * if the access routine is global, we need synchronizations;
   * cf the "dosync" mechanism in the parallel routine. The main
   * difficulty here is that the traversal of the tree has to
   * be level-by-level instead of a postorder. Before and after every
   * level, the process participates to global calls to "elements".
   *
   * Algorithm (ignoring the random vector expansion mechanism):
   * Step1: Rr and Rc are two n x d random matrices.
   * Step2: Sr and Sc are two n x d samples (resp. row and column) of A.
   * Step3: postorder traversal of the HSS tree; at each node:
   *   a/ Extract a local sample.
   *   b/ Compress with interpolative decomposition.
   *   c/ Update random vectors and indices.
   *
   * The way the root is processed might differ:
   *   - If the tree is part of a larger tree, then the root node should
   *     be processed as any other node (i.e., do a/ b/ c/).
   *   - Otherwise, there is nothing to do (because the local samples
   *     would be zero).
   * Parameter "doRoot" decides whether the root node should be processed
   * as any other node (!=0) or not (0).
   * This is useful when calling this routine as part of a parallel
   * compression for example. doRoot also influences the adaptive
   * sampling mechanism; when doRoot!=0, if there is a failure, the
   * compression simply exits.
   *
   * Limitation: for now the algorithm assumes that the HSS tree is binary.
   */

  int n;
  int d, dupd;
  int i, j, k;
  int ii;
  int ib, ie, br, bc;
  int clean;
  int desc[BLACSCTXTSIZE];
  int ctxt;
  int myid, np;
  int curlevel, nblevels;
  int *level_first, *level_last, *traversal;
  int *tmpI, *tmpJ;
  T *tmp;
  HSS_node<T,S> *child1, *child2;
  double t;
  MPI_Request request;

#if defined(levelbylevelseq)
  /* Sequential level by level traversal */
  std::cout << "level by level traversal\n" << std::endl;
  dosync=true;
  comm=MPI_COMM_WORLD;
  for(i=0;i<dim_tree;i++) {
    nodes[i]->nb_calls_prev=0;
    nodes[i]->nb_calls_next=0;
  }
#endif

  /* Some initializations */
  n=a_dim;
  *r=0;
  d=dmin;
  time_Rand=0;
  time_GEMM=0;
  time_ID=0;
  np=1;
  if(dosync) {
    MPI_Comm_rank(comm,&myid);
    MPI_Comm_size(comm,&np);
  }
  if(!doRoot) {
    if(np!=1)
      std::cout<<"np!=1 && !doRoot should not happen!" << std::endl;
    comm=MPI_COMM_WORLD;
    for(i=0;i<dim_tree;i++) {
      nodes[i]->nb_calls_prev=0;
      nodes[i]->nb_calls_next=0;
    }
  }

  /* Interaction with the element access routine (parallel interface) */
  if(elements!=NULL) {
    if(!doRoot) {
      blacs_get_(&IZERO,&IZERO,&ctxt);
      blacs_gridinit_(&ctxt,"R",&IONE,&IONE);
    } else {
      ctxt=nodes[dim_tree-1]->context;
    }
  }
  if(dosync) {
    /* Compute level for each node */
    nblevels=0;
    nodes[dim_tree-1]->level=0;
    for(i=dim_tree-1;i>=0;i--)
      if(nodes[i]->nr_children!=0) {
        nodes[i]->children[0]->level=nodes[i]->level+1;
        nodes[i]->children[1]->level=nodes[i]->level+1;
        nblevels=std::max(nblevels,nodes[i]->children[0]->level);
      }
    nblevels++; /* Levels are 0-based */
    /* For each level, compute the first ("leftmost") and
     * the last ("rightmost).
     */
    level_first=new int[nblevels];
    level_last=new int[nblevels];
    for(j=0;j<nblevels;j++) {
      level_first[j]=dim_tree; /* Larger than last index */
      level_last[j]=-1; /* Smaller than first index */
    }
    for(i=0;i<dim_tree;i++) {
      level_first[nodes[i]->level]=std::min(level_first[nodes[i]->level],i);
      level_last[nodes[i]->level]=std::max(level_last[nodes[i]->level],i);
    }
    /* Level by level traversal (bottom-up, and following the PO induced
     * ordering within levels).
     */
    traversal=new int[dim_tree];
    k=0;
    for(j=nblevels-1;j>=0;j--)
      for(i=0;i<dim_tree;i++)
        if(nodes[i]->level==j)
          traversal[k++]=i;
#if defined(debug)
    printMatrix("level_first:"+SSTR(myid),level_first,1,nblevels,1);
    printMatrix("level_last:"+SSTR(myid),level_last,1,nblevels,1);
    printMatrix("traversal:"+SSTR(myid),traversal,1,dim_tree,1);
#endif

  }

  if(Rr==NULL && Rc==NULL && Sr==NULL && Sc==NULL) {
    clean=1;
    rgen= new StrumpackRand(time(NULL));
    /* Step 1: generating random vectors Rr and Rc */
    Rr=new T[n*d];
    t=MPI_Wtime();
    denseRand(Rr,n,d,rgen);
    time_Rand+=MPI_Wtime()-t;

    Rc=new T[n*d];
    t=MPI_Wtime();
    denseRand(Rc,n,d,rgen);
    time_Rand+=MPI_Wtime()-t;

#if defined(debug)
  printMatrix("Rr",Rr,n,d,n);
  printMatrix("Rc",Rc,n,d,n);
#endif

    /* Step 2: generating samples: Sr=A*Rr, Sc=A'*Rc */
    Sr=new T[n*d];
    t=MPI_Wtime();
    gemm('N','N',n,d,n,ONE,A,n,Rr,n,ZERO,Sr,n);
    time_GEMM+=MPI_Wtime()-t;
    Sc=new T[n*d];
    t=MPI_Wtime();
    gemm('C','N',n,d,n,ONE,A,n,Rc,n,ZERO,Sc,n);
    time_GEMM+=MPI_Wtime()-t;

#if defined(debug)
  printMatrix("Sr",Sr,n,d,n);
  printMatrix("Sc",Sc,n,d,n);
#endif
  } else if (Rr!=NULL && Rc!=NULL && Sr!=NULL && Sc!=NULL) {
    /* Nothing to do, random vectors and samples are provided */
    clean=0;
  } else {
    std::cout << "Error; Rr,Rc,Sr,Sc must be either all NULL or all non-NULL." << std::endl;
    exit(1);
  }

  if(!doRoot)
    /* Start fresh */
    for(i=0;i<dim_tree;i++)
      nodes[i]->state=UNTOUCHED;

  /* Step 3: postorder traversal */
  curlevel=nblevels-1;
  for(ii=0;ii<dim_tree;ii++) {
    if(dosync)
      i=traversal[ii];
    else
      i=ii;
#if defined(debug)
    std::cout << "Node " << i << "; state = " << nodes[i]->state << std::endl;
#endif

    /* If we are in a synchronized traversal, we have to wait for other
     * subtrees and parallel nodes before starting the level.
     */
    if(dosync)
      if(level_first[curlevel]==i)
        participate_to_calls(elements,nodes[i]->nb_calls_prev,comm);

    ib=nodes[i]->tree_ib-1;
    ie=nodes[i]->tree_ie-1;
    /* a/ Extract a local sample */
    if(nodes[i]->state==UNTOUCHED)
      dupd=d;
    else
      dupd=dinc;
    if(nodes[i]->nr_children==0) {
      /* Leaf: Sr_loc = Sr(ib:ie,:)-A(ib:ie,ib:ie) *Rr(ib:ie,:)
       *       Sc_loc = Sc(ib:ie,:)-A(ib:ie,ib:ie)'*Rc(ib:ie,:)
       *       with ib,ie the bounds of the clusters at leaf level.
       * Sr_loc and Sc_loc overwrite Sr(ib:ie,:) and Sc(ib:ie,:) */
      br=ie-ib+1;
      bc=br;
      if(nodes[i]->state==UNTOUCHED) {
        if(A!=NULL) {
          /* Explicit matrix */
          extractSubmatrix(&A[ib*n+ib],br,br,n,NULL,-1,NULL,-1,&(nodes[i]->D));
        } else {
          /* User-given element access */
          nodes[i]->D=new T[br*br];
          tmpI=new int[br];
          for(j=0;j<br;j++)
            tmpI[j]=nodes[i]->tree_ib+j+offset;
          tmpJ=tmpI;
          descset_(desc,&br,&br,&br,&br,&IZERO,&IZERO,&ctxt,&br);
          if(dosync) {
            for(j=0;j<np;j++) {
              if(j!=myid) {
#if defined(debug)
                std::cout << "dosync: " << myid << " sending information to " << j << std::endl;
#endif
                MPI_Isend((void*)desc,BLACSCTXTSIZE,MPI_INTEGER,j,2,comm,&request);
                MPI_Isend((void*)tmpI,br,MPI_INTEGER,j,2,comm,&request);
                MPI_Isend((void*)tmpJ,br,MPI_INTEGER,j,2,comm,&request);
              }
            }
          }
          elements(obj,tmpI,tmpJ,nodes[i]->D,desc);
          if(dosync)
            MPI_Barrier(comm);
          delete[] tmpI;
        }
      }
      if(i==dim_tree-1 && !doRoot) {
        nodes[i]->state=COMPRESSED;
        break;
      }

      gemm('N','N',br,dupd,br,NONE,nodes[i]->D,br,&Rr[ib+(d-dupd)*n],n,ONE,&Sr[ib+(d-dupd)*n],n);
      gemm('C','N',br,dupd,br,NONE,nodes[i]->D,br,&Rc[ib+(d-dupd)*n],n,ONE,&Sc[ib+(d-dupd)*n],n);
    } else if(nodes[i]->nr_children==2) {
      /* Non-leaf: Sr_loc= [Sr_1-B12 *Rr_2;Sr_2-B21 *Rr_1]
       *           Sc_loc= [Sc_1-B21'*Rc_2;Sc_2-B12'*Rc_1]
       *           with B12=A(Ir_1,Ic_2) and B21=A(Ir_2,Ic_1)
       *           with _1 and _2 the children of i.
       * Sr_loc and Sc_loc overwrite Sr(ib:ib+br-1,:) and Sc(ib:ib+br-1,:) */
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];
      br=child1->u_rank+child2->u_rank;
      bc=child1->v_rank+child2->v_rank;
      if(ib!=child1->tree_ib-1 || ie!=child2->tree_ie-1 || child1->tree_ie+1!=child2->tree_ib) {
        std::cout << "Problem with clustering indices." << std::endl;
        exit(1);
      }

      if(nodes[i]->state==UNTOUCHED) {
        if(A!=NULL) {
          /* Explicit matrix */
          extractSubmatrix(A,n,n,n,child1->Ir,child1->u_rank,child2->Ic,child2->v_rank,&nodes[i]->B12);
          extractSubmatrix(A,n,n,n,child2->Ir,child2->u_rank,child1->Ic,child1->v_rank,&nodes[i]->B21);
        } else {
          /* User-given element access */
          nodes[i]->B12=new T[child1->u_rank*child2->v_rank];
          tmpI=new int[child1->u_rank];
          tmpJ=new int[child2->v_rank];
          for(j=0;j<child1->u_rank;j++)
            tmpI[j]=child1->Ir[j]+offset;
          for(j=0;j<child2->v_rank;j++)
            tmpJ[j]=child2->Ic[j]+offset;
          descset_(desc,&child1->u_rank,&child2->v_rank,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&child1->u_rank);
          if(dosync) {
            for(j=0;j<np;j++) {
              if(j!=myid) {
#if defined(debug)
                std::cout << "dosync: " << myid << " sending information to " << j << std::endl;
#endif
                MPI_Isend((void*)desc,BLACSCTXTSIZE,MPI_INTEGER,j,2,comm,&request);
                MPI_Isend((void*)tmpI,child1->u_rank,MPI_INTEGER,j,2,comm,&request);
                MPI_Isend((void*)tmpJ,child2->v_rank,MPI_INTEGER,j,2,comm,&request);
              }
            }
          }
          elements(obj,tmpI,tmpJ,nodes[i]->B12,desc);
          if(dosync)
            MPI_Barrier(comm);
          delete[] tmpI;
          delete[] tmpJ;

          tmpI=new int[child2->u_rank];
          tmpJ=new int[child1->v_rank];
          for(j=0;j<child2->u_rank;j++)
            tmpI[j]=child2->Ir[j]+offset;
          for(j=0;j<child1->v_rank;j++)
            tmpJ[j]=child1->Ic[j]+offset;
          nodes[i]->B21=new T[child2->u_rank*child1->v_rank];
          descset_(desc,&child2->u_rank,&child1->v_rank,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&child2->u_rank);
          if(dosync) {
            for(j=0;j<np;j++) {
              if(j!=myid) {
#if defined(debug)
                std::cout << "dosync: " << myid << " sending information to " << j << std::endl;
#endif
                MPI_Isend((void*)desc,BLACSCTXTSIZE,MPI_INTEGER,j,2,comm,&request);
                MPI_Isend((void*)tmpI,child2->u_rank,MPI_INTEGER,j,2,comm,&request);
                MPI_Isend((void*)tmpJ,child1->v_rank,MPI_INTEGER,j,2,comm,&request);
              }
            }
          }
          elements(obj,tmpI,tmpJ,nodes[i]->B21,desc);
          if(dosync)
            MPI_Barrier(comm);
          delete[] tmpI;
          delete[] tmpJ;
        }
      }

      if(i==dim_tree-1 && !doRoot) {
        nodes[i]->state=COMPRESSED;
        break;
      }

      gemm('N','N',child1->u_rank,dupd,child2->v_rank,NONE,nodes[i]->B12,child1->u_rank,&Rr[child2->tree_ib-1+(d-dupd)*n],n,ONE,&Sr[child1->tree_ib-1+(d-dupd)*n],n);
      gemm('N','N',child2->u_rank,dupd,child1->v_rank,NONE,nodes[i]->B21,child2->u_rank,&Rr[child1->tree_ib-1+(d-dupd)*n],n,ONE,&Sr[child2->tree_ib-1+(d-dupd)*n],n);
      /* Move Sr_child2 below Sr_child1 line by line */
      for(k=0;k<child2->u_rank;k++)
        for(j=d-dupd;j<d;j++)
          Sr[n*j+ib+child1->u_rank+k]=Sr[n*j+child2->tree_ib-1+k];

      gemm('C','N',child1->v_rank,dupd,child2->u_rank,NONE,nodes[i]->B21,child2->u_rank,&Rc[child2->tree_ib-1+(d-dupd)*n],n,ONE,&Sc[child1->tree_ib-1+(d-dupd)*n],n);
      gemm('C','N',child2->v_rank,dupd,child1->u_rank,NONE,nodes[i]->B12,child1->u_rank,&Rc[child1->tree_ib-1+(d-dupd)*n],n,ONE,&Sc[child2->tree_ib-1+(d-dupd)*n],n);
      /* Move Sc_child2 below Sc_child1 line by line */
      for(k=0;k<child2->v_rank;k++)
        for(j=d-dupd;j<d;j++)
          Sc[n*j+ib+child1->v_rank+k]=Sc[n*j+child2->tree_ib-1+k];
    } else {
        std::cout << "Nodes can only have 0 or 2 children. Exiting.\n";
        return;
    }

    /* b/ Compress with Interpolative Decomposition: define the generators U and V
     * and the compressed indices Ir and Ic (resp. row and column indices).
     * Indices in output of ID are local to the matrix passed to the routine;
     * they are shifted in c/
     *
     * Sr/Sc are transposed into tmp before compression.
     *
     * After compression, Sr/Sc=Sr/Sc(Ir/Ic,:).
     */
    if(nodes[i]->state!=COMPRESSED) {
      tmp=new T[d*br];
      copyTrans(br,d,&Sr[ib],n,tmp,d);
      t=MPI_Wtime();
      seq_XInterpol(d,br,tmp,tol,&(nodes[i]->U_E),&(nodes[i]->u_rank),&(nodes[i]->Ir),&(nodes[i]->U_perm));
      time_ID+=MPI_Wtime()-t;
      delete[] tmp;

      if(d-nodes[i]->u_rank<dlim && d+dinc<=dmax) {
        /* Reject this compression; add new random vectors and restart the compression */
        nodes[i]->state=PARTIALLY_COMPRESSED;
        delete[] nodes[i]->U_E;
        delete[] nodes[i]->Ir;
        delete[] nodes[i]->U_perm;
        nodes[i]->u_rank=-1;

        /* If tree is a sequential subtree of a parallel execution */
        if(doRoot)
          break;

        /* Add more columns to Rr and Sr  */
        d+=dinc;
        tmp=new T[n*d]();
        lacpy('N',n,d-dinc,Rr,n,tmp,n);
        delete[] Rr;
        Rr=tmp;
        t=MPI_Wtime();
        denseRand(&Rr[(d-dinc)*n],n,dinc,rgen);
        time_Rand+=MPI_Wtime()-t;
        tmp=new T[n*d]();
        lacpy('N',n,d-dinc,Sr,n,tmp,n);
        delete[] Sr;
        Sr=tmp;
        t=MPI_Wtime();
        gemm('N','N',n,dinc,n,ONE,A,n,&Rr[(d-dinc)*n],n,ZERO,&Sr[(d-dinc)*n],n);
        time_GEMM+=MPI_Wtime()-t;

        /* Add more columns to Rc and Sc  */
        tmp=new T[n*d]();
        lacpy('N',n,d-dinc,Rc,n,tmp,n);
        delete[] Rc;
        Rc=tmp;
        t=MPI_Wtime();
        denseRand(&Rc[(d-dinc)*n],n,dinc,rgen);
        time_Rand+=MPI_Wtime()-t;
        tmp=new T[n*d]();
        lacpy('N',n,d-dinc,Sc,n,tmp,n);
        delete[] Sc;
        Sc=tmp;
        t=MPI_Wtime();
        gemm('C','N',n,dinc,n,ONE,A,n,&Rc[(d-dinc)*n],n,ZERO,&Sc[(d-dinc)*n],n);
        time_GEMM+=MPI_Wtime()-t;

        /* Restart */
        ii=-1;
        continue;
      }

      tmp=new T[d*bc];
      copyTrans(bc,d,&Sc[ib],n,tmp,d);
      t=MPI_Wtime();
      seq_XInterpol(d,bc,tmp,tol,&(nodes[i]->V_E),&(nodes[i]->v_rank),&(nodes[i]->Ic),&(nodes[i]->V_perm));
      time_ID+=MPI_Wtime()-t;
      delete[] tmp;

      if(d-nodes[i]->v_rank<dlim && d+dinc<=dmax) {
        /* Reject this compression; add new random vectors and restart the compression */
        nodes[i]->state=PARTIALLY_COMPRESSED;
        delete[] nodes[i]->V_E;
        delete[] nodes[i]->Ic;
        delete[] nodes[i]->V_perm;
        nodes[i]->v_rank=-1;
        /* Also delete U and recompute it later (should help stability) */
        delete[] nodes[i]->U_E;
        delete[] nodes[i]->Ir;
        delete[] nodes[i]->U_perm;
        nodes[i]->u_rank=-1;

        /* If tree is a sequential subtree of a parallel execution */
        if(doRoot)
          break;

        /* Add more columns to Rr and Sr  */
        d+=dinc;
        tmp=new T[n*d]();
        lacpy('N',n,d-dinc,Rr,n,tmp,n);
        delete[] Rr;
        Rr=tmp;
        t=MPI_Wtime();
        denseRand(&Rr[(d-dinc)*n],n,dinc,rgen);
        time_Rand+=MPI_Wtime()-t;
        tmp=new T[n*d]();
        lacpy('N',n,d-dinc,Sr,n,tmp,n);
        delete[] Sr;
        Sr=tmp;
        t=MPI_Wtime();
        gemm('N','N',n,dinc,n,ONE,A,n,&Rr[(d-dinc)*n],n,ZERO,&Sr[(d-dinc)*n],n);
        time_GEMM+=MPI_Wtime()-t;

        /* Add more columns to Rc and Sc  */
        tmp=new T[n*d]();
        lacpy('N',n,d-dinc,Rc,n,tmp,n);
        delete[] Rc;
        Rc=tmp;
        t=MPI_Wtime();
        denseRand(&Rc[(d-dinc)*n],n,dinc,rgen);
        time_Rand+=MPI_Wtime()-t;
        tmp=new T[n*d]();
        lacpy('N',n,d-dinc,Sc,n,tmp,n);
        delete[] Sc;
        Sc=tmp;
        t=MPI_Wtime();
        gemm('C','N',n,dinc,n,ONE,A,n,&Rc[(d-dinc)*n],n,ZERO,&Sc[(d-dinc)*n],n);
        time_GEMM+=MPI_Wtime()-t;

        /* Restart */
        ii=-1;
        continue;
      }

      *r=std::max(nodes[i]->u_rank,*r);
      nodes[i]->u_rows=br-nodes[i]->u_rank;
      tmp=new T[nodes[i]->u_rows*nodes[i]->u_rank];
      copyTrans(nodes[i]->u_rank,nodes[i]->u_rows,nodes[i]->U_E,nodes[i]->u_rank,tmp,nodes[i]->u_rows);
      delete[] nodes[i]->U_E;
      nodes[i]->U_E=tmp;
      extractSubmatrix(&Sr[ib],nodes[i]->u_rank,d,n,nodes[i]->Ir,nodes[i]->u_rank,NULL,-1,&tmp);
      lacpy('N',nodes[i]->u_rank,d,tmp,nodes[i]->u_rank,&Sr[ib],n);
      delete[] tmp;
      nodes[i]->Jr=new int[nodes[i]->u_rank];
      for(j=0;j<nodes[i]->u_rank;j++)
        nodes[i]->Jr[j]=nodes[i]->Ir[j];

      *r=std::max(nodes[i]->v_rank,*r);
      nodes[i]->v_rows=bc-nodes[i]->v_rank;
      tmp=new T[nodes[i]->v_rows*nodes[i]->v_rank];
      copyTrans(nodes[i]->v_rank,nodes[i]->v_rows,nodes[i]->V_E,nodes[i]->v_rank,tmp,nodes[i]->v_rows);
      delete[] nodes[i]->V_E;
      nodes[i]->V_E=tmp;
      extractSubmatrix(&Sc[ib],nodes[i]->v_rank,d,n,nodes[i]->Ic,nodes[i]->v_rank,NULL,-1,&tmp);
      lacpy('N',nodes[i]->v_rank,d,tmp,nodes[i]->v_rank,&Sc[ib],n);
      delete[] tmp;
      nodes[i]->Jc=new int[nodes[i]->v_rank];
      for(j=0;j<nodes[i]->v_rank;j++)
        nodes[i]->Jc[j]=nodes[i]->Ic[j];
    } else {
      /* An ancestor of i restarted the traversal. We update the new columns of Sr/Sc; we
       * have to be careful because Ir/Ic have been previously transformed to absolute indices.
       * We use Jr/Jc instead. */
      extractSubmatrix(&Sr[ib+(d-dinc)*n],nodes[i]->u_rank,dinc,n,nodes[i]->Jr,nodes[i]->u_rank,NULL,-1,&tmp);
      lacpy('N',nodes[i]->u_rank,dinc,tmp,nodes[i]->u_rank,&Sr[ib+(d-dinc)*n],n);
      delete[] tmp;

      extractSubmatrix(&Sc[ib+(d-dinc)*n],nodes[i]->v_rank,dinc,n,nodes[i]->Jc,nodes[i]->v_rank,NULL,-1,&tmp);
      lacpy('N',nodes[i]->v_rank,dinc,tmp,nodes[i]->v_rank,&Sc[ib+(d-dinc)*n],n);
      delete[] tmp;
    }

    /* c/ Update random vectors and shift indices */
    if(nodes[i]->state==COMPRESSED)
      dupd=dinc;
    else
      dupd=d;
    if(nodes[i]->nr_children==0) {
      /* Leaf: Rr_tau=V'*Rr(ib:ie,:)
       *       Rc_tau=U'*Rc(ib:ie,:)
       *       Ir_tau=Ir_tau(Jr) (with Jr stored in Ir_tau already)
       *       Ic_tau=Ic_tau(Jc) (with Jc stored in Ic_tau already)
       * Rr_tau and Rc_tau overwrite Rr and Rc */
      nodes[i]->seq_applyGenerator('C','V',dupd,ONE,&Rr[ib+(d-dupd)*n],n,ZERO,&Rr[ib+(d-dupd)*n],n);
      if(nodes[i]->state!=COMPRESSED)
        for(j=0;j<nodes[i]->u_rank;j++)
          nodes[i]->Ir[j]+=ib;
      nodes[i]->seq_applyGenerator('C','U',dupd,ONE,&Rc[ib+(d-dupd)*n],n,ZERO,&Rc[ib+(d-dupd)*n],n);
      if(nodes[i]->state!=COMPRESSED)
        for(j=0;j<nodes[i]->v_rank;j++)
          nodes[i]->Ic[j]+=ib;
    } else {
      /* Non-leaf: Rr_tau=V'*[Rr_1; Rr_2]
       *           Rc_tau=U'*[Rc_1; Rc_2]
       *           Ir_tau=[Ir_1 Ir_2](Jr)
       *           Ic_tau=[Ic_1 Ic_2](Jc)
       */
      tmp=new T[bc*dupd];
      lacpy('N',child1->v_rank,dupd,&Rr[child1->tree_ib-1+(d-dupd)*n],n,tmp,bc);
      lacpy('N',child2->v_rank,dupd,&Rr[child2->tree_ib-1+(d-dupd)*n],n,&tmp[child1->v_rank],bc);
      nodes[i]->seq_applyGenerator('C','V',dupd,ONE,tmp,bc,ZERO,&Rr[ib+(d-dupd)*n],n);
      delete[] tmp;
      if(nodes[i]->state!=COMPRESSED)
        for(j=0;j<nodes[i]->u_rank;j++)
          if(nodes[i]->Ir[j]<=child1->u_rank)
            nodes[i]->Ir[j]=child1->Ir[nodes[i]->Ir[j]-1];
          else
            nodes[i]->Ir[j]=child2->Ir[nodes[i]->Ir[j]-child1->u_rank-1];

      tmp=new T[br*dupd];
      lacpy('N',child1->u_rank,dupd,&Rc[child1->tree_ib-1+(d-dupd)*n],n,tmp,br);
      lacpy('N',child2->u_rank,dupd,&Rc[child2->tree_ib-1+(d-dupd)*n],n,&tmp[child1->u_rank],br);
      nodes[i]->seq_applyGenerator('C','U',dupd,ONE,tmp,br,ZERO,&Rc[ib+(d-dupd)*n],n);
      delete[] tmp;
      if(nodes[i]->state!=COMPRESSED)
        for(j=0;j<nodes[i]->v_rank;j++)
          if(nodes[i]->Ic[j]<=child1->v_rank)
            nodes[i]->Ic[j]=child1->Ic[nodes[i]->Ic[j]-1];
          else
            nodes[i]->Ic[j]=child2->Ic[nodes[i]->Ic[j]-child1->v_rank-1];
    }
    nodes[i]->numrand=d;
    nodes[i]->state=COMPRESSED;

    /* If we are in a synchronized traversal, we have to wait for other
     * subtrees and parallel nodes after finishing the level.
     */
    if(dosync)
      if(level_last[curlevel]==i) {
        participate_to_calls(elements,nodes[i]->nb_calls_next,comm);
        curlevel--;
      }

#if defined(debug)
    std::cout<< "u_rank=" << nodes[i]->u_rank <<  "; u_rows=" << nodes[i]->u_rows << std::endl;
    std::cout<< "v_rank=" << nodes[i]->v_rank <<  "; v_rows=" << nodes[i]->v_rows << std::endl;
    if(nodes[i]->nr_children==0)
      printMatrix("D",nodes[i]->D,br,bc,br);
    else {
      printMatrix("B12",nodes[i]->B12,child1->u_rank,child2->v_rank,child1->u_rank);
      printMatrix("B21",nodes[i]->B21,child2->u_rank,child1->v_rank,child2->u_rank);
    }
    printMatrix("U_E",nodes[i]->U_E,nodes[i]->u_rows,nodes[i]->u_rank,nodes[i]->u_rows);
    printMatrix("U_P",nodes[i]->U_perm,1,nodes[i]->u_rank+nodes[i]->u_rows,1);
    printMatrix("Ir",nodes[i]->Ir,1,nodes[i]->u_rank,1);
    printMatrix("V_E",nodes[i]->V_E,nodes[i]->v_rows,nodes[i]->v_rank,nodes[i]->v_rows);
    printMatrix("V_P",nodes[i]->V_perm,1,nodes[i]->v_rank+nodes[i]->v_rows,1);
    printMatrix("Ic",nodes[i]->Ic,1,nodes[i]->v_rank,1);
    printMatrix("Rr",&Rr[ib],nodes[i]->v_rank,d,n);
    printMatrix("Sr",&Sr[ib],nodes[i]->u_rank,d,n);
    printMatrix("Rc",&Rc[ib],nodes[i]->u_rank,d,n);
    printMatrix("Sc",&Sc[ib],nodes[i]->v_rank,d,n);
#endif
  }

#if defined(debug)
  if(!doRoot) {
    i=dim_tree-1;
    ib=nodes[i]->tree_ib-1;
    ie=nodes[i]->tree_ie-1;
    br=ie-ib+1;
    if(nodes[i]->nr_children==0)
      printMatrix("D",nodes[i]->D,br,br,br);
    else {
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];
      printMatrix("B12",nodes[i]->B12,child1->u_rank,child2->v_rank,child1->u_rank);
      printMatrix("B21",nodes[i]->B21,child2->u_rank,child1->v_rank,child2->u_rank);
    }
  }
#endif

  *deff=d;

  if(clean) {
    /* Delete only random vector and samples if they were not provided */
    delete[] Rr;
    delete[] Rc;
    delete[] Sr;
    delete[] Sc;
  }

  if(dosync) {
    delete[] level_first;
    delete[] level_last;
    delete[] traversal;
  }

}

#endif // HSSR_SEQ_COMPRESS_HPP
