/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_NODE_HPP
#define HSS_NODE_HPP

/* Node of an HSS tree; each node has the following:
 *  - A subset tree_ib:tree_ie of 1:N.
 *
 *  - A row generator U of size (u_rank+u_rows) x u_rank, associated with
 *    a set of indices Ir (size u_rank) of the original matrix.
 *    U and Ir come from the interpolative decomposition of a sample
 *    Sr: Sr=Sr(Ir,:)*U; u_rank is the approximate numerical rank of Sr.
 *    The generator U has a special structure and is represented as
 *      U = P * [I ; E]
 *    where: P is a permutation matrix stored in vector U_perm.
 *           I is the u_rank x u_rank identity matrix.
 *           E is u_rows x u_rank stored in U_E.
 *    Jr and Jc are local indices coming from the interpolative decompostion.
 *    They are useful for the adaptive rank mechanism.
 *
 *  - A column generator V of size (v_rank+v_rows) x v_rank, associated with
 *    a set of indices Ic (size v_rank) of the original matrix.
 *    V and Ic come from the interpolative decomposition of a sample
 *    Sc: Sc=Sc(Ic,:)*V; v_rank is the approximate numerical rank of Sc.
 *    The generator V has a special structure and is represented as
 *      V = P * [I ; E]
 *    where: P is a permutation matrix stored in vector V_perm.
 *           I is the v_rank x v_rank identity matrix.
 *           E is v_rows x v_rank stored in V_E.
 *
 *  - D is defined for leaves only: D=A(tree_ib:tree_ie,tree_ib:tree_ie).
 *
 *  - Fields for the ULV factors:
 *      - Vt of size (u_rank+u_rows) x v_rank.
 *      - W of size (u_rank+u_rows) x (u_rank+u_rows).
 *      - Q of size (u_rank+u_rows) x (u_rank+u_rows).
 *      - L of size u_rows x u_rows.
 *        (expect at root: (u_rank+u_rows) x (u_rank+u_rows)).
 *      - D for non-leaves is temporary.
 *      - Dt is temporary.
 *
 *  - B12 and B21 are defined for interior nodes only: B12=A(Ir_1,Ic_2)
 *    and B21=A(Ir_2,Ic_1) with _1 and _2 the children of the node.
 *
 *  - A list of children nodes (size nr_children).
 *
 *  - A list of processes procs[0]...procs[1].
 *
 *  - nprow rows in the associated BLACS grid.
 *
 *  - npcol columns in the associated BLACS grid.
 *
 *  - myrow: row coordinate of a process (LOCAL data).
 *
 *  - mycol: column coordinate of a process (LOCAL data).
 *
 *  - context: BLACS context that represents the *grid* on which
 *    the node is mapped. This is a subset of the of the set of
 *    *processes* on which the node is mapped, because of way
 *    the grid is defined (e.g., 5 processes but a 2x2 grid).
 *
 *  - allctxt: BLACS context that contains all the processes on which
 *    the node is mapped. This is useful for some redistribution
 *    operations.
 *
 *  - isParLeaf: true if the node is a "parallel leaf" (see HSS tree
 *               and HSSR_par_Compress).
 *
 *  - state: either UNTOUCHED, PARTIALLY_COMPRESSED, or COMPRESSED
 *    (cf. compression routines).
 *
 *  - numrand: number of random vectors used for compression.
 *
 * NB: permutations U_perm and V_perm are represented as a bunch of
 *    swaps; U/V_perm[i]=j means that at step i in the RRQR step of
 *    ID we selected column j (i<=j<=n) as the pivot. Conceptually
 *    U_perm and V_perm consist of u_rank and v_rank swaps, but
 *    because of implementation issues they can contain a few
 *    extra swaps so we leave them with size u/v_rows.
 */

template<typename T, typename S> class HSS_node {
public:
  T *U_E;
  int *U_perm;
  int u_rows;
  int u_rank;
  int *descU;
  int *Ir;
  int *Jr;

  T *V_E;
  int *V_perm;
  int v_rows;
  int v_rank;
  int *descV;
  int *Ic;
  int *Jc;

  T *D;
  int tree_ib;
  int tree_ie;

  T *B12;
  T *B21;

  int *piv;
  T *L;
  int *descL;
  T *Vt;
  int *descVt;
  T *W;
  int *descW;
  T *Q;
  int *descQ;
  T *Dt;

  HSS_node<T,S> **children;
  int nr_children;
  HSS_node<T,S> *parent;

  int procs[2];
  int nprow;
  int npcol;
  int myrow;
  int mycol;
  int context;
  int allctxt;
  int isParLeaf;
  int state;
  int numrand;
  int level;
  int nb_calls_prev;
  int nb_calls_next;

  /* Constants */
  static T ZERO;
  static T  ONE;

  void seq_applyGenerator(char, char, int, T, T *, int, T, T *, int);
  void par_applyGenerator(char, char, int, T, T *, int *, T, T *, int *, gridinfo *);

  ~HSS_node();
};

template<>   double HSS_node<  double,double>::ZERO= 0;
template<>   double HSS_node<  double,double>:: ONE= 1;
template<>    float HSS_node<   float, float>::ZERO= 0;
template<>    float HSS_node<   float, float>:: ONE= 1;
template<> dcomplex HSS_node<dcomplex,double>::ZERO= 0;
template<> dcomplex HSS_node<dcomplex,double>:: ONE= 1;
template<> scomplex HSS_node<scomplex, float>::ZERO= 0;
template<> scomplex HSS_node<scomplex, float>:: ONE= 1;

template<typename T, typename S> HSS_node<T,S>::~HSS_node(void) {
  delete[] U_E; U_E=NULL;
  delete[] U_perm; U_perm=NULL;
  delete[] descU; descU=NULL;
  delete[] Ir; Ir=NULL;
  delete[] Jr; Jr=NULL;

  delete[] V_E; V_E=NULL;
  delete[] V_perm; V_perm=NULL;
  delete[] descV; descV=NULL;
  delete[] Ic; Ic=NULL;
  delete[] Jc; Jc=NULL;

  delete[] D; D=NULL;

  delete[] B12; B12=NULL;
  delete[] B21; B21=NULL;

  delete[] piv; piv=NULL;
  delete[] L; L=NULL;
  delete[] Vt; Vt=NULL;
  delete[] W; W=NULL;
  delete[] Q; Q=NULL;

  delete[] children; children=NULL;
}

template<typename T, typename S> void HSS_node<T,S>::seq_applyGenerator(char Trans, char UorV, int d, T alpha, T *B, int ldb, T beta, T *C, int ldc) {
  /* Compute C = beta*C + alpha*X*B with X= U or U^T or V or V^T, where B and C have d columns.
   * The generator X has the special structure X = P * [I ; E], so C = beta*C + alpha*X*B is computed as:
   *   1/ Copy alpha*B into tmp(1:r,:) (tmp has same size as C)
   *   2/ Compute alpha*E*B into tmp(r+1,:)
   *   3/ Apply permutation P (series of swaps with U/V_perm)
   *   4/ C = beta*C+tmp
   *
   * C = beta*C + alpha*X^T*B is computed as:
   *   1/ Copy B into tmp (tmp has same size as B)
   *   2/ Apply permutation P^T (series of swaps with U/V_perm)
   *   3/ tmp(1:r,:) <- alpha*tmp(1:r,:)+alpha*E^T*tmp(r+1:end,:)
   *   4/ C = beta*C+tmp(1:r)
   *
   */

  T *X, *tmp;
  int *P;
  int i, b, r, j;

  if(UorV=='U') {
    X=U_E;
    P=U_perm;
    b=u_rank+u_rows;
    r=u_rank;
  } else {
    X=V_E;
    P=V_perm;
    b=v_rank+v_rows;
    r=v_rank;
  }

  if(Trans=='N') {
    /* 1/ Copy alpha*B into tmp(1:r,:) (tmp has same size as C) */
    tmp=new T[b*d];
    for(j=0;j<d;j++)
      for(i=0;i<r;i++)
        tmp[i+b*j]=alpha*B[i+ldb*j];
    /* 2/ Compute alpha*E*B into tmp(r+1,:) */
    if(b-r>0)
      gemm('N','N',b-r,d,r,alpha,X,b-r,B,ldb,ZERO,&tmp[r],b);
    /* 3/ Apply permutation P (series of swaps with U/V_perm) */
    for(i=b-1;i>=0;i--)
      if(i!=P[i]-1)
        swap(d,&tmp[i],b,&tmp[P[i]-1],b);
    /* 4/ C = beta*C+tmp */
    for(j=0;j<d;j++)
      for(i=0;i<b;i++)
        C[i+ldc*j]=beta*C[i+ldc*j]+tmp[i+b*j];
    delete[] tmp;
  } else {
    /* 1/ Copy B into tmp (tmp has same size as B) */
    tmp=new T[b*d];
    lacpy('N',b,d,B,ldb,tmp,b);
    /* 2/ Apply permutation P^T (series of swaps with U/V_perm) */
    for(i=0;i<b;i++)
      if(i!=P[i]-1)
        swap(d,&tmp[i],b,&tmp[P[i]-1],b);
    /* 3/ tmp(1:r,:) <- alpha*tmp(1:r,:)+alpha*E^T*tmp(r+1:end,:) */
    if(b-r>0)
      gemm('C','N',r,d,b-r,alpha,X,b-r,&tmp[r],b,alpha,tmp,b);
    else
      for(j=0;j<d;j++)
        for(i=0;i<b;i++)
          tmp[i+b*j]=alpha*tmp[i+b*j];
    /* 4/ C = beta*C+tmp(1:r) */
    for(j=0;j<d;j++)
      for(i=0;i<r;i++)
        C[i+ldc*j]=beta*C[i+ldc*j]+tmp[i+b*j];
    delete[] tmp;
  }
}

template<typename T, typename S> void HSS_node<T,S>::par_applyGenerator(char Trans, char UorV, int d, T alpha, T *B, int *descB, T beta, T *C, int * descC, gridinfo *ginfo) {
  /* Compute C = beta*C + alpha*X*B with X= U or U^T or V or V^T, where B and C have d columns.
   * The generator X has the special structure X = P * [I ; E], so C = beta*C + alpha*X*B is computed as:
   *   1/ Copy alpha*B into tmp(1:r,:) (tmp has same size as C)
   *   2/ Compute alpha*E*B into tmp(r+1,:)
   *   3/ Apply permutation P (series of swaps with U/V_perm)
   *   4/ C = beta*C+tmp
   *
   * C = beta*C + alpha*X^T*B is computed as:
   *   1/ Copy B into tmp (tmp has same size as B)
   *   2/ Apply permutation P^T (series of swaps with U/V_perm)
   *   3/ tmp(1:r,:) <- alpha*tmp(1:r,:)+alpha*E^T*tmp(r+1:end,:)
   *   4/ C = beta*C+tmp(1:r)
   *
   */

  T *X, *tmp;
  int *P;
  int i, b, r, ierr;
  int *descX, desc[BLACSCTXTSIZE];
  int locr, locc, idummy;
  int ctxt, nprow, npcol, myrow, mycol, mb, nb;

  ctxt=ginfo->ctxt;
  nprow=ginfo->nprow;
  npcol=ginfo->npcol;
  myrow=ginfo->myrow;
  mycol=ginfo->mycol;
  mb=ginfo->mb;
  nb=ginfo->nb;

  if(UorV=='U') {
    X=U_E;
    descX=descU;
    P=U_perm;
    b=u_rank+u_rows;
    r=u_rank;
  } else {
    X=V_E;
    descX=descV;
    P=V_perm;
    b=v_rank+v_rows;
    r=v_rank;
  }

  if(Trans=='N') {
    /* 1/ Copy alpha*B into tmp(1:r,:) (tmp has same size as C) */
    locr=numroc_(&b,&mb,&myrow,&IZERO,&nprow);
    locc=numroc_(&d,&nb,&mycol,&IZERO,&npcol);
    idummy=std::max(1,locr);
    descinit_(desc,&b,&d,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
    tmp=new T[locr*locc];
    pgeadd('N',r,d,alpha,B,IONE,IONE,descB,ZERO,tmp,IONE,IONE,desc);
    /* 2/ Compute alpha*E*B into tmp(r+1,:) */
    pgemm('N','N',b-r,d,r,alpha,X,IONE,IONE,descX,B,IONE,IONE,descB,ZERO,tmp,r+1,IONE,desc);
    /* 3/ Apply permutation P (series of swaps with U/V_perm) */
    for(i=b;i>=1;i--)
      if(i!=P[i-1])
        pswap(d,tmp,i,IONE,desc,b,tmp,P[i-1],IONE,desc,b);
    /* 4/ C = beta*C+tmp */
    pgeadd('N',b,d,ONE,tmp,IONE,IONE,desc,beta,C,IONE,IONE,descC);
    delete[] tmp;
  } else {
    /* 1/ Copy B into tmp (tmp has same size as B) */
    locr=numroc_(&b,&mb,&myrow,&IZERO,&nprow);
    locc=numroc_(&d,&nb,&mycol,&IZERO,&npcol);
    idummy=std::max(1,locr);
    descinit_(desc,&b,&d,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
    tmp=new T[locr*locc];
    pgeadd('N',b,d,ONE,B,IONE,IONE,descB,ZERO,tmp,IONE,IONE,desc);
    /* 2/ Apply permutation P^T (series of swaps with U/V_perm) */
    for(i=1;i<=b;i++)
      if(i!=P[i-1])
        pswap(d,tmp,i,IONE,desc,b,tmp,P[i-1],IONE,desc,b);
    /* 3/ tmp(1:r,:) <- alpha*tmp(1:r,:)+alpha*E^T*tmp(r+1:end,:) */
    pgemm('C','N',r,d,b-r,alpha,X,IONE,IONE,descX,tmp,r+1,IONE,desc,alpha,tmp,IONE,IONE,desc);
    /* 4/ C = beta*C+tmp(1:r) */
    pgeadd('N',r,d,ONE,tmp,IONE,IONE,desc,beta,C,IONE,IONE,descC);
    delete[] tmp;
  }
}

#endif // HSS_NODE_HPP
