/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_PAR_SOLVE_HPP
#define HSS_PAR_SOLVE_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSS_par_Solve(T *X, int *descX, int d, T *B, int * descB, gridinfo *ginfo, int partial) {
  /* Solution of A*X=B using the ULV factors of A.
   * B and X have d columns.
   * "partial":
   *   - If 0, do the entire solve.
   *   - If 1, do only the forward pass.
   *   - If 2, do only the backward pass.
   */

  int i, j, k, br, bc, br1, br2, ib;
  int mb, nb;
  int myid, np;
  int istart, iend;
  int info;
  int last;
  int ldl, idummy, ierr;
  int locr, locc;
  T *Xl, *Bl;
  T *tmp, *yt, *xi, *y12, *z12;
  T **f, **x, **z;
  int *children1, *children2;
  int ch1, ch2;
  int *descBX, *descf, *descx, *descy, *descz, *desc;
  int descyc[BLACSCTXTSIZE],descyt[BLACSCTXTSIZE], descyk[BLACSCTXTSIZE];
  gridinfo tmpinfo, tmpinfo1, tmpinfo2;
  int proc_b, proc_g, proc_e, proc_bc, proc_gc;
  HSS_node<T,S> *child1, *child2;

  mb=ginfo->mb;
  nb=ginfo->nb;
  MPI_Comm_rank(comm,&myid);
  MPI_Comm_size(comm,&np);

  /* Here f, x, y, are not fields of a node but arrays indexed following the
   * postorder; the pointers "children" of a node are not enough to find the
   * right in y and z, what we need is numbers in the postorder. "children1"
   * and "children2" are the same information as children but they provide
   * numbers (indices in the tree).
   *
   * children1: children[i]==j: node number j is left child of node number i.
   * child1: pointer to left child (type HSS_node) of node nodes[i].
   *
   * When partial==1: the "y" structure (field of the HSS tree) is saved at
   * the end of the forward pass (bottom-up) in order to be reused in the
   * backward pass (partial==2). With partial=1, [z1;z2] at the root node
   * is saved into zk12 (field of the tree). This is useful for RHS
   * reduction.
   */

  children1=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children1[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[0]) {
          children1[i]=j;
          break;
        }
  }
  children2=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children2[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[1]) {
          children2[i]=j;
          break;
        }
  }

  if(partial) {
    /* Traverse only the subtree rooted at the left child
     * of the root node.
     */
    last=knode;
  } else {
    /* Traverse the whole tree. */
    last=dim_tree-1;
  }

  /* Beginning of forward solve */
  if(partial!=2) {
    /* As in the compression routine, the first thing to do is to redistribute
     * the input following the clustering induced by the parallel leaves of
     * the tree.
     */
    distVectorTree(B,descB,d,&Bl,&descBX,ginfo,nodes[dim_tree-1]->allctxt);

    /* Step 1: postorder traversal (except root) */
    f=new T*[dim_tree];
    for(i=0;i<dim_tree;i++)
      f[i]=NULL;
    y=new T*[dim_tree];
    for(i=0;i<dim_tree;i++)
      y[i]=NULL;
    z=new T*[dim_tree];
    for(i=0;i<dim_tree;i++)
      z[i]=NULL;

    /* Serial subtree */
    if(subtree[2*myid]>=0 && (!partial || subtree[2*myid+1]<=knode)) {
      istart=subtree[2*myid];
      iend=subtree[2*myid+1];
      if(partial)
        iend=std::min(knode,iend);
      ldl=np==1?a_dim:nodes[iend]->tree_ie-nodes[iend]->tree_ib+1;
      for(i=istart;i<=iend;i++) {
        ib=nodes[i]->tree_ib-nodes[istart]->tree_ib;
        if(nodes[i]->nr_children==0) {
          br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
          bc=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        } else {
          child1=nodes[i]->children[0];
          child2=nodes[i]->children[1];
          br=child1->u_rank+child2->u_rank;
          bc=child1->v_rank+child2->v_rank;
          br1=child1->u_rows+child1->u_rank;
          br2=child2->u_rows+child2->u_rank;
          ch1=children1[i];
          ch2=children2[i];
        }

        if(nodes[i]->nr_children!=0) {
          /* f{i} = [ ft{child 1}_2 - W{child 1}_2*Q{child 1}_1'*y{child 1} - B12*z{child 2} ]
           *        [ ft{child 2}_2 - B12*z{child 1} - W{child 2}_2*Q{child 2}_1'*y{child 2} ]
           *      = [ (1,1) - (1,2) - (1,3) ]
           *        [ (2,1) - (2,2) - (2,3) ]
           */
          f[i]=new T[br*d];

          /* (1,1) */
          lacpy('N',child1->u_rank,d,&f[ch1][child1->u_rows],br1,f[i],br);

          /* (1,2) if child1->u_rows!=0 (otherwise y{child 1} empty) */
          if(child1->u_rows) {
            /* Q*y */
            tmp=new T[br1*d];
            gemm('C','N',br1,d,child1->u_rows,ONE,child1->Q,br1,y[ch1],child1->u_rows,ZERO,tmp,br1);

            /* W*Q*y */
            gemm('N','N',child1->u_rank,d,br1,NONE,&child1->W[child1->u_rows],br1,tmp,br1,ONE,f[i],br);
            delete[] tmp;
          }

          /* (1,3) if child2->u_rows!=0 or non-leaf (otherwise z{child 2} 0) */
          if(child2->u_rows||child2->nr_children)
            gemm('N','N',child1->u_rank,d,child2->v_rank,NONE,nodes[i]->B12,child1->u_rank,z[ch2],child2->v_rank,ONE,f[i],br);

          /* (2,1) */
          lacpy('N',child2->u_rank,d,&f[ch2][child2->u_rows],br2,&f[i][child1->u_rank],br);

          /* (2,2) if child1->u_rows!=0 or non-leaf (otherwise z{child 1} 0) */
          if(child1->u_rows||child1->nr_children)
            gemm('N','N',child2->u_rank,d,child1->v_rank,NONE,nodes[i]->B21,child2->u_rank,z[ch1],child1->v_rank,ONE,&f[i][child1->u_rank],br);

          /* (2,3): if child2->u_rows!=0 (otherwise y{child 2} empty) */
          if(child2->u_rows) {
            /* Q*y */
            tmp=new T[br2*d];
            gemm('C','N',br2,d,child2->u_rows,ONE,child2->Q,br2,y[ch2],child2->u_rows,ZERO,tmp,br2);

            /* W*Q*y */
            gemm('N','N',child2->u_rank,d,br2,NONE,&child2->W[child2->u_rows],br2,tmp,br2,ONE,&f[i][child1->u_rank],br);
            delete[] tmp;
          }

        } else {
          f[i]=new T[br*d];
          lacpy('N',br,d,&Bl[ib],ldl,f[i],br);
        }

        if(i==last) {
          /* Triangular solve with traditional LU factors at root */
          info=getrs('N',br,d,nodes[i]->L,br,nodes[i]->piv,f[i],br);
          if(info!=0)
            std::cout << "Something went wrong with triangular solve at root node" << std::endl;
          y[i]=new T[br*d];
          lacpy('N',br,d,f[i],br,y[i],br);
          if(partial==1 && nodes[i]->nr_children) {
            /* Save [z1;z2] into zk12 (field of the tree); will be used for RHS reduction */
            zk12=new T[bc*d];
            lacpy('N',child1->v_rank,d,z[ch1],child1->v_rank,zk12,bc);
            lacpy('N',child2->v_rank,d,z[ch2],child2->v_rank,&zk12[child1->v_rank],bc);
            desc_zk12=new int[BLACSCTXTSIZE];
            descset_(desc_zk12,&bc,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&bc);
          }
        } else {

          /* ft{i} = [-U_E I ; 0 I ] * U_perm' * f{i} */
          /* f{i} is overwritten */
          for(k=0;k<br;k++)
            if(k!=nodes[i]->U_perm[k]-1)
              swap(d,&f[i][k],br,&f[i][nodes[i]->U_perm[k]-1],br);

          if(nodes[i]->u_rows) {
            /* Otherwise ft{i} = U_perm' * f{i} empty, y{i} empty */

            /* ft{i} = [ft{i}_1 ; ft{i}_2] with ft{i}_1=f{i}(u_rank+1:end,:)-U_E*f{i}(1:u_rank,:) and ft{i}_2=f{i}(1:u_rank,:) */
            tmp=new T[br*d]();
            lacpy('N',nodes[i]->u_rows,d,&f[i][nodes[i]->u_rank],br,tmp,br);
            gemm('N','N',nodes[i]->u_rows,d,nodes[i]->u_rank,NONE,nodes[i]->U_E,nodes[i]->u_rows,f[i],br,ONE,tmp,br);
            lacpy('N',nodes[i]->u_rank,d,f[i],br,&tmp[nodes[i]->u_rows],br);
            delete[] f[i];
            f[i]=tmp;

            /* y{i} = L{i}\ft{i}_1 */
            y[i]=new T[nodes[i]->u_rows*d];
            lacpy('N',nodes[i]->u_rows,d,f[i],br,y[i],nodes[i]->u_rows);
            trsm('L','L','N','N',nodes[i]->u_rows,d,ONE,nodes[i]->L,nodes[i]->u_rows,y[i],nodes[i]->u_rows);
          }
          /* z{i} = Vt{i}_1' * y{i} + V{i}' * [z{child 1} ; z{child 2}] (non leaves) */
          z[i]=new T[nodes[i]->v_rank*d]();
          if(nodes[i]->u_rows)
            gemm('C','N',nodes[i]->v_rank,d,nodes[i]->u_rows,ONE,nodes[i]->Vt,br,y[i],nodes[i]->u_rows,ZERO,z[i],nodes[i]->v_rank);
          if(nodes[i]->nr_children!=0) {
            tmp=new T[bc*d];
            lacpy('N',child1->v_rank,d,z[ch1],child1->v_rank,tmp,bc);
            lacpy('N',child2->v_rank,d,z[ch2],child2->v_rank,&tmp[child1->v_rank],bc);
            nodes[i]->seq_applyGenerator('C','V',d,ONE,tmp,bc,ONE,z[i],nodes[i]->v_rank);
            delete[] tmp;
          }
        }
      }
    }

    /* Second, parallel nodes are visited. This is the same algorithm as
     * in the lower part, except that arrays are constantly distributed
     * in a 2D block-cyclic format and parallel Interpolative Decomposition
     * is used.
     */
    for(j=0;j<par_lvl+1;j++) {
      /* Same loop for leaves and interior nodes */
      if(j>=1) {
        /* Follow the chain of parallel interior nodes */
        i=par_nodes[j-1];
        if(i>last)
          break;
      } else {
        /* Find local parallel leaf (if any) */
        for(i=0;i<=last;i++) {
          proc_b=nodes[i]->procs[0];
          proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
          proc_e=nodes[i]->procs[1];
          if(proc_e>proc_b && myid>=proc_b && myid <= proc_e && nodes[i]->nr_children==0)
            /* Leaf mapped on several processes on which myid works */
            break;
        }
        if(i>last)
         /* No parallel leaf, go to interior nodes */
         continue;
      }

      /* Processing node i */
      if(nodes[i]->nr_children==0) {
        br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        bc=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      } else {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        br=child1->u_rank+child2->u_rank;
        bc=child1->v_rank+child2->v_rank;
        br1=child1->u_rows+child1->u_rank;
        br2=child2->u_rows+child2->u_rank;
        ch1=children1[i];
        ch2=children2[i];
      }

      /* Node i is mapped on proc_b...proc_e but only
       * proc_b...proc_g are active (in the grid).
       */
      proc_b=nodes[i]->procs[0];
      proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
      proc_e=nodes[i]->procs[1];
      desc=new int[BLACSCTXTSIZE];
      y12=NULL;
      z12=NULL;
      descf=NULL;
      descy=NULL;
      descz=NULL;

      if(nodes[i]->nr_children!=0) {
        /* f{i} = [ ft{child 1}_2 - W{child 1}_2*Q{child 1}_1'*y{child 1} - B12*z{child 2} ]
         *        [ ft{child 2}_2 - B12*z{child 1} - W{child 2}_2*Q{child 2}_1'*y{child 2} ]
         *      = [ (1,1) - (1,2) - (1,3) ]
         *        [ (2,1) - (2,2) - (2,3) ]
         */

        tmpinfo.ctxt=nodes[i]->context;
        tmpinfo.nprow=nodes[i]->nprow;
        tmpinfo.npcol=nodes[i]->npcol;
        tmpinfo.myrow=nodes[i]->myrow;
        tmpinfo.mycol=nodes[i]->mycol;
        tmpinfo.mb=ginfo->mb;
        tmpinfo.nb=ginfo->nb;

        /* (1,1) and (2,1): [ft{child1}_2 ; ft{child2}_2] using mergeVector */
        if(myid<child1->procs[0]+child1->nprow*child1->npcol) {
          /* Mapped on child 1 */
          tmp=f[ch1];
          idummy=std::max(1,numroc_(&br1,&mb,&child1->myrow,&IZERO,&child1->nprow));
          descinit_(desc,&br1,&d,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
        } else if(myid>=child2->procs[0] && myid<child2->procs[0]+child2->nprow*child2->npcol) {
          /* Mapped on child 2 */
          tmp=f[ch2];
          idummy=std::max(1,numroc_(&br2,&mb,&child2->myrow,&IZERO,&child2->nprow));
          descinit_(desc,&br2,&d,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
        } else {
          /* Out of both grids */
          tmp=NULL;
        }

        mergeVector(tmp,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->u_rank,child1->u_rows+1,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->u_rank,child2->u_rows+1,proc_b,proc_g,d,desc,&f[i],&descf,nodes[i]->allctxt,&tmpinfo,comm);

        /* (1,3) and (2,2): [z{child1} ; z{child2}] using mergeVector */
        if(myid<child1->procs[0]+child1->nprow*child1->npcol) {
          /* Mapped on child 1 */
          tmp=z[ch1];
          idummy=std::max(1,numroc_(&child1->v_rank,&mb,&child1->myrow,&IZERO,&child1->nprow));
          descinit_(desc,&child1->v_rank,&d,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
        } else if(myid>=child2->procs[0] && myid<child2->procs[0]+child2->nprow*child2->npcol) {
          /* Mapped on child 2 */
          tmp=z[ch2];
          idummy=std::max(1,numroc_(&child2->v_rank,&mb,&child2->myrow,&IZERO,&child2->nprow));
          descinit_(desc,&child2->v_rank,&d,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
        } else {
          /* Out of both grids */
          tmp=NULL;
        }

        mergeVector(tmp,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->v_rank,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->v_rank,IONE,proc_b,proc_g,d,desc,&z12,&descz,nodes[i]->allctxt,&tmpinfo,comm);

        if(myid>=proc_b && myid<=proc_g) {
          /* (1,3) if child2->u_rows!=0 or non-leaf (otherwise z{child 2} 0) */
          if(child2->u_rows||child2->nr_children) {
            idummy=std::max(1,numroc_(&child1->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
            descinit_(desc,&child1->u_rank,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
            pgemm('N','N',child1->u_rank,d,child2->v_rank,NONE,nodes[i]->B12,IONE,IONE,desc,z12,child1->v_rank+1,IONE,descz,ONE,f[i],IONE,IONE,descf);
          }

          /* (2,2) if child1->u_rows!=0 or non-leaf (otherwise z{child 1} 0) */
          if(child1->u_rows||child1->nr_children) {
            idummy=std::max(1,numroc_(&child2->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
            descinit_(desc,&child2->u_rank,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
            pgemm('N','N',child2->u_rank,d,child1->v_rank,NONE,nodes[i]->B21,IONE,IONE,desc,z12,IONE,IONE,descz,ONE,f[i],child1->u_rank+1,IONE,descf);

          }
        }

        /* (1,2) and (2,3): [yt1 ; yt2] using mergeVector */
        if(myid<child1->procs[0]+child1->nprow*child1->npcol) {
          /* Mapped on child 1: compute W{child 1}_2*Q{child 1}_1'*y{child 1} */
          locr=numroc_(&child1->u_rank,&mb,&child1->myrow,&IZERO,&child1->nprow);
          locc=numroc_(&d,&nb,&child1->mycol,&IZERO,&child1->npcol);
          yt=new T[locr*locc]();
          idummy=std::max(1,locr);
          descinit_(descyt,&child1->u_rank,&d,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
          if(child1->u_rows) {
            /* Otherwise y empty and yt=0 */
            /* Q*y */
            locr=numroc_(&br1,&mb,&child1->myrow,&IZERO,&child1->nprow);
            locc=numroc_(&d,&nb,&child1->mycol,&IZERO,&child1->npcol);
            tmp=new T[locr*locc];
            idummy=std::max(1,locr);
            descinit_(desc,&br1,&d,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
            idummy=std::max(1,numroc_(&child1->u_rows,&mb,&child1->myrow,&IZERO,&child1->nprow));
            descinit_(descyc,&child1->u_rows,&d,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
            if(child1->nprow*child1->npcol==1) {
              /* We need to set descQ */
              idummy=std::max(1,numroc_(&br1,&mb,&child1->myrow,&IZERO,&child1->nprow));
              child1->descQ=new int[BLACSCTXTSIZE];
              descinit_(child1->descQ,&br1,&br1,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
            }
            pgemm('C','N',br1,d,child1->u_rows,ONE,child1->Q,IONE,IONE,child1->descQ,y[ch1],IONE,IONE,descyc,ZERO,tmp,IONE,IONE,desc);
            /* W*Q*y */
            if(child1->nprow*child1->npcol==1) {
              /* We need to set descW */
              idummy=std::max(1,numroc_(&br1,&mb,&child1->myrow,&IZERO,&child1->nprow));
              child1->descW=new int[BLACSCTXTSIZE];
              descinit_(child1->descW,&br1,&br1,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
            }
            pgemm('N','N',child1->u_rank,d,br1,ONE,child1->W,child1->u_rows+1,IONE,child1->descW,tmp,IONE,IONE,desc,ZERO,yt,IONE,IONE,descyt);
            delete[] tmp;
          }
        } else if(myid>=child2->procs[0] && myid<child2->procs[0]+child2->nprow*child2->npcol) {
          /* Mapped on child 2: compute W{child 2}_2*Q{child 2}_1'*y{child 2} */
          locr=numroc_(&child2->u_rank,&mb,&child2->myrow,&IZERO,&child2->nprow);
          locc=numroc_(&d,&nb,&child2->mycol,&IZERO,&child2->npcol);
          yt=new T[locr*locc]();
          idummy=std::max(1,locr);
          descinit_(descyt,&child2->u_rank,&d,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
          if(child2->u_rows) {
            /* Otherwise y empty and yt=0 */
            /* Q*y */
            locr=numroc_(&br2,&mb,&child2->myrow,&IZERO,&child2->nprow);
            locc=numroc_(&d,&nb,&child2->mycol,&IZERO,&child2->npcol);
            tmp=new T[locr*locc];
            idummy=std::max(1,locr);
            descinit_(desc,&br2,&d,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
            idummy=std::max(1,numroc_(&child2->u_rows,&mb,&child2->myrow,&IZERO,&child2->nprow));
            descinit_(descyc,&child2->u_rows,&d,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
            if(child2->nprow*child2->npcol==1) {
              /* We need to set descQ */
              idummy=std::max(1,numroc_(&br2,&mb,&child2->myrow,&IZERO,&child2->nprow));
              child2->descQ=new int[BLACSCTXTSIZE];
              descinit_(child2->descQ,&br2,&br2,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
            }
            pgemm('C','N',br2,d,child2->u_rows,ONE,child2->Q,IONE,IONE,child2->descQ,y[ch2],IONE,IONE,descyc,ZERO,tmp,IONE,IONE,desc);
            /* W*Q*y */
            if(child2->nprow*child2->npcol==1) {
              /* We need to set descW */
              idummy=std::max(1,numroc_(&br2,&mb,&child2->myrow,&IZERO,&child2->nprow));
              child2->descW=new int[BLACSCTXTSIZE];
              descinit_(child2->descW,&br2,&br2,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
            }
            pgemm('N','N',child2->u_rank,d,br2,ONE,child2->W,child2->u_rows+1,IONE,child2->descW,tmp,IONE,IONE,desc,ZERO,yt,IONE,IONE,descyt);
            delete[] tmp;
          }
        } else {
          /* Out of both grids */
          yt=NULL;
        }

        mergeVector(yt,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->u_rank,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->u_rank,IONE,proc_b,proc_g,d,descyt,&y12,&descy,nodes[i]->allctxt,&tmpinfo,comm);

        if(myid>=proc_b && myid<=proc_g) {
          pgeadd('N',br,d,NONE,y12,IONE,IONE,descy,ONE,f[i],IONE,IONE,descf);
        }

        if(!(partial && last))
          delete[] descz;
        delete[] y12;
        delete[] descy;
      } else {
        /* Parallel leaf: copy Bl into f[i] */
        if(myid>=proc_b && myid<=proc_g) {
          locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          f[i]=new T[locr*locc];
          idummy=std::max(1,locr);
          descf=new int[BLACSCTXTSIZE];
          descinit_(descf,&br,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          placpy('N',br,d,Bl,IONE,IONE,descBX,f[i],IONE,IONE,descf);
        }
      }

      if(myid>=proc_b && myid<=proc_g) {
        if(i==last) {
          /* Triangular solve with traditional LU factors at root */
          locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          y[i]=new T[locr*locc];
          placpy('N',br,d,f[i],IONE,IONE,descf,y[i],IONE,IONE,descf);
          ierr=pgetrs('N',br,d,nodes[i]->L,IONE,IONE,nodes[i]->descL,nodes[i]->piv,y[i],IONE,IONE,descf);
          copyDesc(descf,descyk);
          if(ierr!=0)
            std::cout << "Something went wrong with triangular solve at root node" << std::endl;
          if(partial==1 && nodes[i]->nr_children) {
            /* Save [z1;z2] into zk12 (field of the tree); will be used for RHS reduction */
            desc_zk12=new int[BLACSCTXTSIZE];
            copyDesc(descz,desc_zk12);
            locr=numroc_(&bc,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
            locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
            zk12=new T[locr*locc];
            placpy('N',bc,d,z12,IONE,IONE,descz,zk12,IONE,IONE,desc_zk12);
          }
        } else {
          /* ft{i} = [-U_E I ; 0 I ] * U_perm' * f{i} */
          /* f{i} is overwritten */
          for(k=1;k<=br;k++)
            if(k!=nodes[i]->U_perm[k-1])
              pswap(d,f[i],k,IONE,descf,br,f[i],nodes[i]->U_perm[k-1],IONE,descf,br);

          if(nodes[i]->u_rows) {
            /* Otherwise ft{i} = U_perm' * f{i}, y{i} empty */
            locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
            locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
            tmp=new T[locr*locc];

            /* ft{i} = [ft{i}_1 ; ft{i}_2] with ft{i}_1=f{i}(u_rank+1:end,:)-U_E*f{i}(1:u_rank,:) and ft{i}_2=f{i}(1:u_rank,:) */
            pgeadd('N',nodes[i]->u_rows,d,ONE,f[i],nodes[i]->u_rank+1,IONE,descf,ZERO,tmp,IONE,IONE,descf);
            pgemm('N','N',nodes[i]->u_rows,d,nodes[i]->u_rank,NONE,nodes[i]->U_E,IONE,IONE,nodes[i]->descU,f[i],IONE,IONE,descf,ONE,tmp,IONE,IONE,descf);
            pgeadd('N',nodes[i]->u_rank,d,ONE,f[i],IONE,IONE,descf,ZERO,tmp,nodes[i]->u_rows+1,IONE,descf);
            delete[] f[i];
            f[i]=tmp;

            /* y{i} = L{i}\ft{i}_1 */
            locr=numroc_(&nodes[i]->u_rows,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
            locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
            y[i]=new T[locr*locc];
            idummy=std::max(1,locr);
            descy=new int[BLACSCTXTSIZE];
            descinit_(descy,&nodes[i]->u_rows,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
            pgeadd('N',nodes[i]->u_rows,d,ONE,f[i],IONE,IONE,descf,ZERO,y[i],IONE,IONE,descy);
            ptrsm('L','L','N','N',nodes[i]->u_rows,d,ONE,nodes[i]->L,IONE,IONE,nodes[i]->descL,y[i],IONE,IONE,descy);
          }

          /* z{i} = Vt{i}_1' * y{i} + V{i}' * [z{child 1} ; z{child 2}] (non leaves) */
          locr=numroc_(&nodes[i]->v_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          z[i]=new T[locr*locc]();
          idummy=std::max(1,locr);
          descz=new int[BLACSCTXTSIZE];
          descinit_(descz,&nodes[i]->v_rank,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          if(nodes[i]->u_rows) {
            pgemm('C','N',nodes[i]->v_rank,d,nodes[i]->u_rows,ONE,nodes[i]->Vt,IONE,IONE,nodes[i]->descVt,y[i],IONE,IONE,descy,ZERO,z[i],IONE,IONE,descz);
            delete[] descy;
          }
          if(nodes[i]->nr_children!=0) {
            idummy=std::max(1,numroc_(&bc,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
            descinit_(desc,&bc,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
            nodes[i]->par_applyGenerator('C','V',d,ONE,z12,desc,ONE,z[i],descz,&tmpinfo);

          }
          delete[] descz;
        }
      }

      delete[] desc;
      delete[] descf;
      delete[] z12;

    }
    if(partial) {
      /* We need to move y[knode] (descyk) from the knode-grid to the root grid.
       * We make a copy in yr (desc_yr) (fields of the tree).
       */
      i=dim_tree-1;
      proc_b=nodes[i]->procs[0];
      proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
      child1=nodes[knode];
      br=child1->u_rank+child1->u_rows;
      proc_bc=child1->procs[0];
      proc_gc=child1->procs[0]+child1->nprow*child1->npcol-1;

      if(myid>=proc_bc && myid<=proc_gc) {
        locr=numroc_(&br,&mb,&child1->myrow,&IZERO,&child1->nprow);
        idummy=std::max(1,locr);
        descinit_(descyk,&br,&d,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
      } else {
        descset_(descyk,&br,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }
      if(myid>=proc_b && myid<=proc_g) {
        locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        idummy=std::max(1,locr);
        yr=new T[locr*locc];
        desc_yr=new int[BLACSCTXTSIZE];
        descinit_(desc_yr,&br,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      } else {
        yr=NULL;
        desc_yr=new int[BLACSCTXTSIZE];
        descset_(desc_yr,&br,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }
      pgemr2d(br,d,y[knode],IONE,IONE,descyk,yr,IONE,IONE,desc_yr,nodes[i]->allctxt);
    }

    for(i=0;i<dim_tree-1;i++) {
      delete[] f[i];
      delete[] z[i];
    }
    delete[] f;
    delete[] z;

    delete[] Bl;

  }
  /* End of forward solve */

  /* Beginning of backward solve */
  if(partial!=1) {
    /* Redistribute the solution vector along the tree */
    distVectorTree(X,descX,d,&Xl,&descBX,ginfo,nodes[dim_tree-1]->allctxt);

    x=new T*[dim_tree];
    for(i=0;i<dim_tree;i++)
      x[i]=NULL;

    if(partial) {
      /* We need to move yr (desc_yr) from the root grid back to the
       * knode-grid in y[i].
       */
      i=dim_tree-1;
      proc_b=nodes[i]->procs[0];
      proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
      child1=nodes[knode];
      br=child1->u_rank+child1->u_rows;
      proc_bc=child1->procs[0];
      proc_gc=child1->procs[0]+child1->nprow*child1->npcol-1;

      if(myid>=proc_bc && myid<=proc_gc) {
        locr=numroc_(&br,&mb,&child1->myrow,&IZERO,&child1->nprow);
        locc=numroc_(&d,&nb,&child1->mycol,&IZERO,&child1->npcol);
        idummy=std::max(1,locr);
        y[i]=new T[locr*locc];
        descy=new int[BLACSCTXTSIZE];
        descinit_(descy,&br,&d,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
      } else {
        y[i]=NULL;
        descy=new int[BLACSCTXTSIZE];
        descset_(descy,&br,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }
      pgemr2d(br,d,yr,IONE,IONE,desc_yr,y[knode],IONE,IONE,descy,nodes[i]->allctxt);
      delete[] descy;
    }

    /* Step 2: reverse postorder traversal */
    for(j=par_lvl;j>=0;j--) {
      /* Same loop for leaves and interior nodes */
      if(j>=1) {
        /* Follow the chain of parallel interior nodes */
        i=par_nodes[j-1];
        if(i>last)
          continue;
      } else {
        /* Find local parallel leaf (if any) */
        for(i=0;i<=last;i++) {
          proc_b=nodes[i]->procs[0];
          proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
          proc_e=nodes[i]->procs[1];
          if(proc_e>proc_b && myid>=proc_b && myid <= proc_g && nodes[i]->nr_children==0)
            /* Leaf mapped on several processes on which myid works */
            break;
        }
        if(i>last)
         /* No parallel leaf, done */
         break;
      }

      /* Node i is mapped on proc_b...proc_e but only
       * proc_b...proc_g are active (in the grid).
       */
      proc_b=nodes[i]->procs[0];
      proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
      proc_e=nodes[i]->procs[1];

      /* Partial solution at root is stored in y */
      if(i==last) {
        if(myid>=proc_b && myid<=proc_g) {
          if(nodes[i]->nr_children!=0) {
            child1=nodes[i]->children[0];
            child2=nodes[i]->children[1];
            br=child1->u_rank+child2->u_rank;
          } else
            br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
          locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          x[i]=new T[locr*locc];
          lacpy('N',locr,locc,y[i],locr,x[i],locr);
        }
      }

      if(nodes[i]->nr_children!=0) {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        br=child1->u_rank+child2->u_rank;
        bc=child1->v_rank+child2->v_rank;
        br1=child1->u_rows+child1->u_rank;
        br2=child2->u_rows+child2->u_rank;
        ch1=children1[i];
        ch2=children2[i];
        descx=new int[BLACSCTXTSIZE];

        /* Split/redistribute x{i} to the children: xi=x{i}_x on processes working on child x */
        if(myid>=proc_b && myid<=proc_g) {
          idummy=std::max(1,numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
          descinit_(descx,&br,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        } else
          descset_(descx,&br,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);

        tmpinfo1.ctxt=child1->context;
        tmpinfo1.nprow=child1->nprow;
        tmpinfo1.npcol=child1->npcol;
        tmpinfo1.myrow=child1->myrow;
        tmpinfo1.mycol=child1->mycol;
        tmpinfo1.mb=ginfo->mb;
        tmpinfo1.nb=ginfo->nb;
        tmpinfo2.ctxt=child2->context;
        tmpinfo2.nprow=child2->nprow;
        tmpinfo2.npcol=child2->npcol;
        tmpinfo2.myrow=child2->myrow;
        tmpinfo2.mycol=child2->mycol;
        tmpinfo2.mb=ginfo->mb;
        tmpinfo2.nb=ginfo->nb;

        splitVector(x[i],child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->u_rank,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->u_rank,d,descx,&xi,&desc,nodes[i]->allctxt,&tmpinfo1,&tmpinfo2,comm);

        if(myid<child1->procs[0]+child1->nprow*child1->npcol) {
          /* Child 1: x{child1} = Q{child1}'*[y{child1};x{i}_1] */
          locr=numroc_(&br1,&mb,&child1->myrow,&IZERO,&child1->nprow);
          locc=numroc_(&d,&nb,&child1->mycol,&IZERO,&child1->npcol);
          x[ch1]=new T[locr*locc];
          idummy=std::max(1,locr);
          descinit_(descx,&br1,&d,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);

          if(child1->u_rows) {
            descy=new int[BLACSCTXTSIZE];
            idummy=std::max(1,numroc_(&child1->u_rows,&mb,&child1->myrow,&IZERO,&child1->nprow));
            descinit_(descy,&child1->u_rows,&d,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
            if(child1->nprow*child1->npcol==1) {
              /* We need to set descQ */
              idummy=std::max(1,numroc_(&br1,&mb,&child1->myrow,&IZERO,&child1->nprow));
              child1->descQ=new int[BLACSCTXTSIZE];
              descinit_(child1->descQ,&br1,&br1,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
            }
            pgemm('C','N',br1,d,child1->u_rows,ONE,child1->Q,IONE,IONE,child1->descQ,y[ch1],IONE,IONE,descy,ZERO,x[ch1],IONE,IONE,descx);
            delete[] descy;
            pgemm('C','N',br1,d,child1->u_rank,ONE,child1->Q,child1->u_rows+1,IONE,child1->descQ,xi,IONE,IONE,desc,ONE,x[ch1],IONE,IONE,descx);
          } else {
          /* Q{child 1} = I (not stored), y{child 1} empty => x{child 1}=x{i}_1 */
            placpy('N',child1->u_rank,d,xi,IONE,IONE,desc,x[ch1],IONE,IONE,descx);
          }
          delete[] descx;

        } else if(myid>=child2->procs[0] && myid<child2->procs[0]+child2->nprow*child2->npcol) {
          /* Child 2: x{child2} = Q{child2}'*[y{child2};x{i}_2] */
          locr=numroc_(&br2,&mb,&child2->myrow,&IZERO,&child2->nprow);
          locc=numroc_(&d,&nb,&child2->mycol,&IZERO,&child2->npcol);
          x[ch2]=new T[locr*locc];
          idummy=std::max(1,locr);
          descinit_(descx,&br2,&d,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);

          if(child2->u_rows) {
            descy=new int[BLACSCTXTSIZE];
            idummy=std::max(1,numroc_(&child2->u_rows,&mb,&child2->myrow,&IZERO,&child2->nprow));
            descinit_(descy,&child2->u_rows,&d,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
            if(child2->nprow*child2->npcol==1) {
              /* We need to set descQ */
              idummy=std::max(1,numroc_(&br2,&mb,&child2->myrow,&IZERO,&child2->nprow));
              child2->descQ=new int[BLACSCTXTSIZE];
              descinit_(child2->descQ,&br2,&br2,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
            }
            pgemm('C','N',br2,d,child2->u_rows,ONE,child2->Q,IONE,IONE,child2->descQ,y[ch2],IONE,IONE,descy,ZERO,x[ch2],IONE,IONE,descx);
            delete[] descy;
            pgemm('C','N',br2,d,child2->u_rank,ONE,child2->Q,child2->u_rows+1,IONE,child2->descQ,xi,IONE,IONE,desc,ONE,x[ch2],IONE,IONE,descx);
          } else {
            /* Q{child 2} = I (not stored), y{child 2} empty => x{child 2}=x{i}_2 */
            placpy('N',child2->u_rank,d,xi,IONE,IONE,desc,x[ch2],IONE,IONE,descx);
          }
          delete[] descx;
        }

        delete[] desc;
      } else {
        if(myid>=proc_b && myid<=proc_g) {
          /* For a parallel leaf i, x[i] and Xl have the same size and distribution */
          br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
          placpy('N',br,d,x[i],IONE,IONE,descBX,Xl,IONE,IONE,descBX);
        }
      }
    }

    /* Serial subtree */
    if(subtree[2*myid]>=0 && (!partial || subtree[2*myid+1]<=knode)) {
      istart=subtree[2*myid];
      iend=subtree[2*myid+1];
      ldl=np==1?a_dim:nodes[iend]->tree_ie-nodes[iend]->tree_ib+1;
      for(i=iend;i>=istart;i--) {
        ib=nodes[i]->tree_ib-nodes[istart]->tree_ib;
        if(nodes[i]->nr_children==0) {
          br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
          bc=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        } else {
          child1=nodes[i]->children[0];
          child2=nodes[i]->children[1];
          br=child1->u_rank+child2->u_rank;
          bc=child1->v_rank+child2->v_rank;
          br1=child1->u_rows+child1->u_rank;
          br2=child2->u_rows+child2->u_rank;
          ch1=children1[i];
          ch2=children2[i];
        }

        /* Partial solution at root is stored in y */
        if(i==last) {
          x[i]=new T[br*d];
          lacpy('N',br,d,y[i],br,x[i],br);
        }

        if(nodes[i]->nr_children!=0) {
          /* Child 1: x{child1} = Q{child1}'*[y{child1};x{i}_1] */
          if(child1->u_rows) {
            tmp=new T[br1*d]();
            lacpy('N',child1->u_rows,d,y[ch1],child1->u_rows,tmp,br1);
            lacpy('N',child1->u_rank,d,x[i],br,&tmp[child1->u_rows],br1);
            x[ch1]=new T[br1*d];
            gemm('C','N',br1,d,br1,ONE,child1->Q,br1,tmp,br1,ZERO,x[ch1],br1);
            delete[] tmp;
          } else {
            /* Q{child 1} = I (not stored), y{child 1} empty => x{child 1}=x{i}_1 */
            x[ch1]=new T[br1*d];
            lacpy('N',br1,d,x[i],br,x[ch1],br1);
          }

          /* Child 2: x{child2} = Q{child2}'*[y{child2};x{i}_2] */
          if(child2->u_rows) {
            tmp=new T[br2*d]();
            lacpy('N',child2->u_rows,d,y[ch2],child2->u_rows,tmp,br2);
            lacpy('N',child2->u_rank,d,&x[i][child1->u_rank],br,&tmp[child2->u_rows],br2);
            x[ch2]=new T[br2*d];
            gemm('C','N',br2,d,br2,ONE,child2->Q,br2,tmp,br2,ZERO,x[ch2],br2);
            delete[] tmp;
          } else {
            /* Q{child 2} = I (not stored), y{child 2} empty => x{child 2}=x{i}_2 */
            x[ch2]=new T[br2*d];
            lacpy('N',br2,d,&x[i][child1->u_rank],br,x[ch2],br2);
          }
        } else {
          /* Copy the piece of solution x[i] to the corresponding part
           * of the final solution X. */
          lacpy('N',br,d,x[i],br,&Xl[ib],ldl);
        }
      }
    }

    /* Redistribute Xl in 2D block-cyclic form */
    undistVectorTree(Xl,descBX,d,X,descX,ginfo,nodes[dim_tree-1]->allctxt,partial);

    delete[] Xl;
    for(i=0;i<dim_tree-1;i++) {
      delete[] x[i];
      delete[] y[i];
    }
    delete[] x;
    delete[] y;
    y=NULL;
  }
  /* End of backward solve */
}

#endif // HSS_PAR_SOLVE_HPP
