/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_PAR_TOOLS_HPP
#define HSS_PAR_TOOLS_HPP

template<typename T, typename S> void HSS_tree<T,S>::distVectorTree(T *X, int *descX, int d, T **Y, int **descY, gridinfo *ginfo, int globctxt) {
  /* Distribution of the rows of an n x d matrix ("vector") according to
   * the HSS tree.
   * The input vector X (n x d) is distributed in a 2D block cyclic format.
   * n is implicitly given by the HSS tree.
   * The output Y is a distributed version of X whose rows conform with
   * the clustering of [1 n] defined by the parallel leaves of the HSS tree.
   * Cf. routine HSSR_par_Compress for details.
   *
   * The algorithm is the following:
   * for each node i in the tree
   *   if i is a parallel leaf
   *     processes that own i initializes a descriptor with i.context
   *     the other processes initialize a descriptor with dummy context -1
   *     pgemr2d redistributes X(i.tree_ib:i.tree_ie,:) to the processes that own i
   *     and stores it in Y.
   */
  int i;
  int myid, np, ierr;
  int mb, nb;
  int n, locr, locc;
  int idummy;
  int *desc;

  MPI_Comm_rank(comm,&myid);
  MPI_Comm_size(comm,&np);

  mb=ginfo->mb;
  nb=ginfo->nb;

  *Y=NULL;
  *descY=NULL;
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->isParLeaf) {
      desc=new int[BLACSCTXTSIZE];
      n=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      if(   nodes[i]->procs[0]<=myid
         && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
        ) {
        locr=numroc_(&n,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        *Y=new T[locr*locc];
        idummy=std::max(1,locr);
        descinit_(desc,&n,&d,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        *descY=desc;
      } else {
        descset_(desc,&n,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }
      pgemr2d(n,d,X,nodes[i]->tree_ib,IONE,descX,*Y,IONE,IONE,desc,globctxt);
    }
  }

}

template<typename T, typename S> void HSS_tree<T,S>::undistVectorTree(T *X, int *descX, int d, T *Y, int *descY, gridinfo *ginfo, int globctxt, int partial) {
  /* The "opposite" of distVectorTree.
   * The input vector X is distributed following the tree, i.e., its
   * rows conform with the clustering of [1 n] defined by the parallel
   * leaves of the HSS tree.
   * The output vector Y is a 2D block-cyclic version of X.
   * Cf. routine HSSR_par_Compress for details.
   *
   * The algorithm is the following:
   * for each node i in the tree
   *   if i is a parallel leaf
   *     processes that own i initializes a descriptor with i.context
   *     the other processes initialize a descriptor with dummy context -1
   *     pgemr2d redistributes X(i.tree_ib:i.tree_ie,:) to all the processes,
   *     at the right place in Y.
   *
   * The only difference with distVectorTree is that Y and descY are assumed
   * to be allowed/set already.
   *
   * If partial=1, traverse only the subtree rooted at the right child of the
   * root node (this is used in par_Product). If partial=2, traverse only the
   * subtree rooted at the left child of the root node (this is used in par_Solve
   * for the partial backward solve / solution expansion).
   */
  int i;
  int first, last;
  int myid, np, ierr;
  int mb, nb;
  int n;
  int idummy;
  int desc[BLACSCTXTSIZE];

  MPI_Comm_rank(comm,&myid);
  MPI_Comm_size(comm,&np);

  mb=ginfo->mb;
  nb=ginfo->nb;

  switch(partial) {
    case 0:
      first=0;
      last=dim_tree-1;
      break;
    case 1:
      first=knode+1;
      last=dim_tree-2;
      break;
    case 2:
      first=0;
      last=knode;
      break;
  }

  for(i=first;i<=last;i++) {
    if(nodes[i]->isParLeaf) {
      n=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      if(   nodes[i]->procs[0]<=myid
         && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
        ) {
        idummy=std::max(1,numroc_(&n,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&n,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      } else {
        descset_(desc,&n,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }
      pgemr2d(n,d,X,IONE,IONE,desc,Y,nodes[i]->tree_ib,IONE,descY,globctxt);
    }
  }

}

template<typename T, typename S> void HSS_tree<T,S>::extractRowVectorTree(T *X, int *descX, int *I, int nI, int d, T *Y, int *descY, gridinfo *ginfo, int globctxt) {
  /* Similar to undistVectorTree, except the output vector Y has only nI rows
   * sorted following I. At each leaf of the tree, X is in 2D block cyclic
   * form and has ie-ib+1 rows from which we extracted selected rows corresponding
   * to I.
   */
  int i, j;
  int myid, np, ierr;
  int mb, nb;
  int n;
  int idummy;
  int desc[BLACSCTXTSIZE];

  MPI_Comm_rank(comm,&myid);
  MPI_Comm_size(comm,&np);

  mb=ginfo->mb;
  nb=ginfo->nb;
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->isParLeaf) {
      n=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      for(j=0;j<nI;j++) {
        if(nodes[i]->tree_ib<=I[j] && I[j]<=nodes[i]->tree_ie) {
          /* Node i owns Y(j,:) (corresponding to A(I[j],:) of the matrix),
           * locally stored in X(I[j]-i.tree_ib,:).
           */
          if(   nodes[i]->procs[0]<=myid
             && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
            ) {
            idummy=std::max(1,numroc_(&n,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
            descinit_(desc,&n,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          } else {
            descset_(desc,&n,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
          }
          pgemr2d(IONE,d,X,I[j]-nodes[i]->tree_ib+1,IONE,desc,Y,j+1,IONE,descY,globctxt);
        }
      }
    }
  }


}

template<typename T, typename S> void HSS_tree<T,S>::distMatrixTree(T *A, int *descA, int par_lvl, T **D, int **descD, T **A12, int **descA12, T **A21, int **descA21, gridinfo *ginfo, int globctxt) {
  /* Distribution of the rows of an n x n matrix according to
   * the HSS tree.
   * The input matrix X (n x n) is distributed in a 2D block cyclic format.
   * n is implicitly given by the HSS tree.
   * The output is:
   *   D: diagonal block corresponding to the parallel leaf that myid owns.
   *   A12 and A21: par_lvl blocks corresponding to off-diagonal blocks of
   *     A for the par_lvl parallel nodes on which myid works.
   * Cf. routine HSSR_par_Compress for details.
   *
   * The algorithm is the following:
   * my_i=-1
   * for each node i in the tree
   *   if i is a parallel leaf
   *     processes that own i initializes a descriptor with i.context
   *     the other processes initialize a descriptor with dummy context -1
   *     pgemr2d redistributes A(i.tree_ib:i.tree_ie,:i.tree_ib:i.tree_ie) to the processes that own i (into D).
   *   if i is a parallel node (and not a leaf of the HSS tree)
   *     processes that own i initializes a descriptor with i.context
   *     the other processes initialize a descriptor with dummy context -1
   *     denote c1 and c2 the two children of i, with their intervals [ib1 ie1] and[ib2 ie2].
   * A(ib1:ie1,ib2:ie2) (resp. A(ib2:ie2,ib1:ie1)).
   *     if myid owns i, my_i++
   *     pgemr2d redistributes A(ib1:ie1,ib2:ie2) to the processes that own i (into A12(my_i)).
   *     pgemr2d redistributes A(ib2:ie2,ib1:ie1) to the processes that own i (into A21(my_i)).
   *
   * NB: here "myid owns i" means that 1/ i is mapped on myid
   *                               AND 2/ myid is in the BLACS grid associated to i.
   * In terms of implementation, i->procs[1] is the last process on which myid
   * is mapped, but i->procs[0]+i->nprow*i->npcol-1 is the last process that *works*
   * on i.
   */
  int i, my_i, idummy;
  int myid, np, ierr;
  int mb, nb;
  int m, n, locr, locc;
  int *desc;
  HSS_node<T,S> *child1, *child2;

  MPI_Comm_rank(comm,&myid);
  MPI_Comm_size(comm,&np);

  mb=ginfo->mb;
  nb=ginfo->nb;

  my_i=-1;
  *D=NULL;
  for(i=0;i<par_lvl;i++) {
    A12[i]=NULL;
    descA12[i]=NULL;
    A21[i]=NULL;
    descA21[i]=NULL;
  }

  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->isParLeaf) {
      /* Distribute diagonal blocks for parallel leaves */
      n=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      desc=new int[BLACSCTXTSIZE];
      if(   nodes[i]->procs[0]<=myid
         && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
        ) {
        locr=numroc_(&n,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&n,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        idummy=std::max(1,locr);
        *D=new T[locr*locc];
        descinit_(desc,&n,&n,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        *descD=desc;
      } else {
        descset_(desc,&n,&n,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }
      pgemr2d(n,n,A,nodes[i]->tree_ib,nodes[i]->tree_ib,descA,*D,IONE,IONE,desc,globctxt);
    }
    else if(nodes[i]->procs[1]>nodes[i]->procs[0] && nodes[i]->nr_children>0) {
      /* Distribute off-diagonal blocks for parallel interior nodes) */
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];
      m=child1->tree_ie-child1->tree_ib+1;
      n=child2->tree_ie-child2->tree_ib+1;
      desc=new int[BLACSCTXTSIZE];
      /* A12 */
      if(   nodes[i]->procs[0]<=myid
         && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
        ) {
        locr=numroc_(&m,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&n,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        idummy=std::max(1,locr);
        my_i++;
        A12[my_i]=new T[locr*locc];
        descinit_(desc,&m,&n,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        descA12[my_i]=desc;
      } else {
        descset_(desc,&m,&n,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
        if(nodes[i]->procs[0]<=myid && nodes[i]->procs[1]>=myid)
          /* If i is mapped on myid but myid is not in the grid,
           * my_i still needs to be incremented so that A12 for
           * myid is not offset (w.r.t. to the processes in the grid
           * (this would cause problem to retrieve A12 at ancestors
           * of this node). */
          my_i++;
      }
      pgemr2d(m,n,A,child1->tree_ib,child2->tree_ib,descA,A12[my_i],IONE,IONE,desc,globctxt);
      /* A21 */
      desc=new int[BLACSCTXTSIZE];
      if(   nodes[i]->procs[0]<=myid
         && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
        ) {
        locr=numroc_(&n,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&m,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        idummy=std::max(1,locr);
        /* my_i already incremented before... */
        A21[my_i]=new T[locr*locc];
        descinit_(desc,&n,&m,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        descA21[my_i]=desc;
      } else {
        descset_(desc,&n,&m,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }
      pgemr2d(n,m,A,child2->tree_ib,child1->tree_ib,descA,A21[my_i],IONE,IONE,desc,globctxt);
    }
  }

}

template<typename T, typename S> void HSS_tree<T,S>::extractLocalSubtree(HSS_tree<T,S> *loc_htree) {
  /* Extract the subtree defined by nodes subtree[2*myid]..subtree[2*myid+1], where
   * subtree[2*myid] is the leftmost and bottommost descendant of subtree[2*myid+1].
   * The indices are shifted by subtree[2*myid]->tree_ib-1, so that the local
   * tree corresponds to a cluster of indices starting at 1.
   *
   * NB: the routine generates a complete binary tree, so if the global tree is not
   * a complete binary tree (e.g., comb tree), the routine may fail.
   */
  int m, n;
  int istart, iend;
  int myid;

  MPI_Comm_rank(comm,&myid);

  istart=subtree[2*myid];
  iend=subtree[2*myid+1];

  if(istart==-1)
    return;

  n=iend-istart+1;
  m=static_cast<int>(floor(log(static_cast<float>(n))/log(2e0)));
  if(pow(2e0,m)-1<n)
    m++;
  if(pow(2e0,m)-1!=n) {
    std::cout << "Internal error in extractLocalSubtree" << std::endl;
    exit(1);
  }

  n=nodes[iend]->tree_ie-nodes[istart]->tree_ib+1;

  loc_htree->generateBinaryTree(m,n);

#if defined(debug)
  loc_htree->print();
#endif

}

template<typename T, typename S> void HSS_tree<T,S>::restart(T **Rr, T **Sr, T** Rc, T **Sc, int **descRS, int n, int d, T* A, int *descA, gridinfo *ginfo, int allctxt, StrumpackRand *r, double *tRand, double *tGEMM) {
  int myid;
  int *desc;
  T *tmp;

  MPI_Comm_rank(comm,&myid);

  /* Generate new random vectors and samples */
  generateSamples(Rr,Sr,Rc,Sc,descRS,n,d,A,descA,ginfo,r,comm,tRand,tGEMM);

  /* Redistribute along the tree */
  distVectorTree(*Rr,*descRS,d,&tmp,&desc,ginfo,allctxt);
  delete[] *Rr;
  *Rr=tmp;

  delete []desc;
  distVectorTree(*Rc,*descRS,d,&tmp,&desc,ginfo,allctxt);
  delete[] *Rc;
  *Rc=tmp;

  delete []desc;
  distVectorTree(*Sr,*descRS,d,&tmp,&desc,ginfo,allctxt);
  delete[] *Sr;
  *Sr=tmp;

  delete []desc;
  distVectorTree(*Sc,*descRS,d,&tmp,&desc,ginfo,allctxt);
  delete[] *Sc;
  *Sc=tmp;
  delete[] *descRS;
  *descRS=desc;
}

#endif // HSS_PAR_TOOLS_HPP
