/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_SEQ_EXTRACTION_HPP
#define HSS_SEQ_EXTRACTION_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSS_seq_Extraction(T *B, int *I, int nI, int *J, int nJ) {
  /* Sequential element extraction: B(i,j) is an approximation to A(I(i),J(j)),
   * obtained from the HSS form of A.
   * B is assumed to be allocated.
   *
   * The algorithm is essentially a matrix-vector with sparse RHS and sparse
   * solution: A(i,j)=(A e_j)(i).
   *
   * We "prune" the tree by computing which elements need to be traversed:
   *  - For each element J(j) of J, the set of nodes that needs to be processed
   *  in the bottom-up traversal of the tree is the path from the unique leaf
   *  that contains J(j) up to the root node.
   *  - For each element I(i) of I, the set of nodes that needs to be processed
   *  in the top-traversal of the tree is the path from the unique leaf that
   *  contains I(i) up to the root node.
   *
   * In the bottom-up traversal, each node has a set of active columns:
   *  - For a leaf, it corresponds to elements of J that are in the index set
   *    of that leaf.
   *  - For non-leaves, it is the union of the sets of the children.
   * In the top-down traversal
   *
   * Algorithm:
   * Step1: postorder traversal:
   *     if i leaf
   *       scan J to compute the number of actives columns at i
   *       if non-zero,
   *         y{i}=V{i}'*[e_j1...e_jnJ]
   *     otherwise, y{i}=V{i}'*[y{child1}; y{child2}]
   *                where y{child1} or y{child2} (but not both) can be 0
   *                if child1 or child2 wasn't traversed.
   *                y{child1} and y{child2} have different column counts;
   *                the merge is actually
   *                [y{child1} 0; 0 y{child2}]
   * Step2: reverse postorder traversal
   *   scan J to compute isneeded: isneeded[i]=1 if i needs to be
   *   traversed
   *   if i needs to be traversed:
   *     if i leaf, B(I,:)=U{i}*z{i}+D{i}*x(I,:)
   *     otherwise,
   *       r1/2: row-rank for child1/2
   *       z{child1}=B12{i}*y{child2}+U{i}(1:r1)*z{i}
   *       z{child2}=B21{i}*y{child1}+U{i}(r1+1:r1+r2)*z{i}
   *       (if root node, the second term in the sum is not here).
   *
   */

  int i, j, k;
  int b;
  int rows_y, rows_y1, rows_y2, rows_z, rows_z1, rows_z2;
  T *tmp;
  T **y, **z;
  int *children1, *children2;
  int *columns;
  bool *active;
  int ch1, ch2;
  HSS_node<T,S> *child1, *child2;

  /* Here y and z are not fields of a node but arrays indexed following the
   * postorder; the pointers "children" of a node are not enough to find the
   * right in y and z, what we need is numbers in the postorder. "children1"
   * and "children2" are the same information as children but they provide
   * numbers (indices in the tree).
   *
   * children1: children[i]==j: node number j is left child of node number i.
   * child1: pointer to left child (type HSS_node) of node nodes[i].
   */

  children1=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children1[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[0]) {
          children1[i]=j;
          break;
        }
  }
  children2=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children2[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[1]) {
          children2[i]=j;
          break;
        }
  }

  /* Step 1: postorder traversal (except root) */
  columns=new int[dim_tree];
  y=new T*[dim_tree-1];
  for(i=0;i<dim_tree-1;i++)
    y[i]=NULL;
  for(i=0;i<dim_tree-1;i++) {
    b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
    rows_y=nodes[i]->v_rank;
    if(nodes[i]->nr_children==0) {
      /* Leaf */
      columns[i]=0;
      for(j=0;j<nJ;j++)
        /* Is J(j) in the index set of i? */
        if(nodes[i]->tree_ib<=J[j] && J[j]<=nodes[i]->tree_ie)
          columns[i]++;
      if(!columns[i])
        continue; /* Node is not active */
      y[i]=new T[rows_y*columns[i]]();
      tmp=new T[b*columns[i]]();
      columns[i]=0;
      for(j=0;j<nJ;j++)
        if(nodes[i]->tree_ib<=J[j] && J[j]<=nodes[i]->tree_ie) {
          tmp[J[j]-nodes[i]->tree_ib+b*columns[i]]=ONE;
          columns[i]++;
        }
      nodes[i]->seq_applyGenerator('C','V',columns[i],ONE,tmp,b,ZERO,y[i],rows_y);
      delete[] tmp;
    } else {
      /* Non-leaf: y{i}=V{i}'*[y{child1} 0;0 y{child2}] */
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];
      ch1=children1[i];
      ch2=children2[i];

      rows_y1=child1->v_rank;
      rows_y2=child2->v_rank;
      columns[i]=columns[ch1]+columns[ch2];
      if(!columns[i])
        continue; /* Node is not active */
      y[i]=new T[rows_y*columns[i]]();
      tmp=new T[(rows_y1+rows_y2)*columns[i]]();

      columns[i]=0;
      columns[ch1]=0;
      columns[ch2]=0;
      for(j=0;j<nJ;j++)
        if(nodes[i]->tree_ib<=J[j] && J[j]<=nodes[i]->tree_ie) {
          if(child1->tree_ib<=J[j] && J[j]<=child1->tree_ie) {
            /* From left child */
            lacpy('N',rows_y1,IONE,&y[ch1][rows_y1*columns[ch1]],rows_y1,&tmp[(rows_y1+rows_y2)*columns[i]],rows_y1+rows_y2);
            columns[i]++;
            columns[ch1]++;
          } else {
            /* From right child */
            lacpy('N',rows_y2,IONE,&y[ch2][rows_y2*columns[ch2]],rows_y2,&tmp[rows_y1+(rows_y1+rows_y2)*columns[i]],rows_y1+rows_y2);
            columns[i]++;
            columns[ch2]++;
          }
        }

      nodes[i]->seq_applyGenerator('C','V',columns[i],ONE,tmp,rows_y1+rows_y2,ZERO,y[i],rows_y);
      delete[] tmp;
    }
  }
  /* Step 2: reverse postorder traversal */
  active=new bool[dim_tree];
  /* First, we determine active nodes using a bottom-up traversal */
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->nr_children==0) {
      active[i]=false;
      for(j=0;j<nI;j++)
        if(nodes[i]->tree_ib<=I[j] && I[j]<=nodes[i]->tree_ie) {
          active[i]=true;
          break;
        }
    } else {
      ch1=children1[i];
      ch2=children2[i];
      active[i]=active[ch1] || active[ch2];
    }
  }
  z=new T*[dim_tree-1];
  for(i=0;i<dim_tree-1;i++)
    z[i]=NULL;
  for(i=dim_tree-1;i>=0;i--) {
    if(!active[i])
      continue; /* Node is not active */
    rows_z=nodes[i]->u_rank;
    if(nodes[i]->nr_children==0) {
      /* Leaf: b(:,:)=U{i}*z{i}+D(:,J).
       * Note that we could compress the rows
       */
      b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      tmp=new T[b*nJ]();
      for(j=0;j<nJ;j++)
        if(nodes[i]->tree_ib<=J[j] && J[j]<=nodes[i]->tree_ie)
          lacpy('N',b,IONE,&nodes[i]->D[(J[j]-nodes[i]->tree_ib)*b],b,&tmp[b*j],b);

      if(i!=dim_tree-1)
        /* Avoid weird case leaf=root (1 node in the tree) */
        nodes[i]->seq_applyGenerator('N','U',nJ,ONE,z[i],rows_z,ONE,tmp,b);

      /* Extract values (rows) from tmp into B */
      for(j=0;j<nI;j++)
        if(nodes[i]->tree_ib<=I[j] && I[j]<=nodes[i]->tree_ie)
          lacpy('N',IONE,nJ,&tmp[I[j]-nodes[i]->tree_ib],b,&B[j],nI);
      delete[] tmp;
    } else {
      /* Non-leaf */
      ch1=children1[i];
      ch2=children2[i];
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];
      rows_z1=child1->u_rank;
      rows_z2=child2->u_rank;
      rows_y1=child1->v_rank;
      rows_y2=child2->v_rank;
      /* Child 1: z{child1}=B12{i}*y{child2}+U{i}(1:r1)*z{i}; first part.
       * We first compute B12*y{child2} and scatter the columns[ch2] columns
       * into nJ columns.
       */
      if(active[ch1]) {
        z[ch1]=new T[rows_z1*nJ]();
        if(columns[ch2]) {
          tmp=new T[rows_z1*columns[ch2]];
          gemm('N','N',rows_z1,columns[ch2],rows_y2,ONE,nodes[i]->B12,child1->u_rank,y[ch2],rows_y2,ZERO,tmp,rows_z1);
          columns[ch2]=0;
          for(j=0;j<nJ;j++)
            if(child2->tree_ib<=J[j] && J[j]<=child2->tree_ie) {
              lacpy('N',rows_z1,IONE,&tmp[rows_z1*columns[ch2]],rows_z1,&z[ch1][rows_z1*j],rows_z1);
              columns[ch2]++;
            }
          delete[] tmp;
        }
      }

      /* Child 2: z{child2}=B21{i}*y{child1}+U{i}(r1+1:r1+r2)*z{i}; first part
       * We first compute B21*y{child1} and scatter the columns[ch1] columns
       * into nJ columns.
       */
      if(active[ch2]) {
        z[ch2]=new T[rows_z2*nJ]();
        if(columns[ch1]) {
          tmp=new T[rows_z2*columns[ch1]];
          gemm('N','N',rows_z2,columns[ch1],rows_y1,ONE,nodes[i]->B21,child2->u_rank,y[ch1],rows_y1,ZERO,tmp,rows_z2);
          columns[ch1]=0;
          for(j=0;j<nJ;j++)
            if(child1->tree_ib<=J[j] && J[j]<=child1->tree_ie) {
              lacpy('N',rows_z2,IONE,&tmp[rows_z2*columns[ch1]],rows_z2,&z[ch2][rows_z2*j],rows_z2);
              columns[ch1]++;
            }
          delete[] tmp;
        }
      }

      /* Child 1 + Child 2; second part */
      if(i!=dim_tree-1) {
        tmp=new T[(rows_z1+rows_z2)*nJ]();
        nodes[i]->seq_applyGenerator('N','U',nJ,ONE,z[i],rows_z,ZERO,tmp,rows_z1+rows_z2);
        for(j=0;j<nJ;j++) {
          if(active[ch1])
            for(k=0;k<rows_z1;k++)
              z[ch1][j*rows_z1+k]+=tmp[j*(rows_z1+rows_z2)+k];
          if(active[ch2])
            for(k=0;k<rows_z2;k++)
              z[ch2][j*rows_z2+k]+=tmp[j*(rows_z1+rows_z2)+k+rows_z1];
        }
        delete[] tmp;
      }
    }
  }

  for(i=0;i<dim_tree-1;i++) {
    delete[] y[i];
    delete[] z[i];
  }
  delete[] y;
  delete[] z;
  delete[] columns;
  delete[] active;

}

#endif // HSS_SEQ_EXTRACTION_HPP
