/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_SEQ_ULV_HPP
#define HSS_SEQ_ULV_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSS_seq_ULV(int doRoot, int partial, int saveVhat) {
  /* ULV-like factorization of an HSS matrix.
   *
   * Parameter "doRoot" decides whether the root node should be processed
   * as any other node (!=0) or not (0).
   * This is useful when calling this routine as part of a parallel
   * factorization for example.
   *
   * When "partial" is nonzero, only the lefmost subtree (rooted at the
   * "left"/"first" child of the root node) is factored. This works only
   * when the tree has 2 levels or more.
   *
   * "saveVhat" is useful when calling this routine on a subtree rooted
   * at the "knode". In that case, the root node of the tree on which
   * this routine is called is an interior node of the whole tree,
   * thus it has U/V generators and we can compute Vhat.
   */

  int i, j, k, br, bc, br1, br2;
  int info;
  int last;
  T *Ve, *tmp, *tau;
  HSS_node<T,S> *child1, *child2;

  if(partial) {
    /* Searching for the index of the left child of the root node. */
    for(last=0;last<dim_tree;last++) {
      if(nodes[last]==nodes[dim_tree-1]->children[0])
        break;
    }
    knode=last;
  } else {
    last=dim_tree-1;
  }

  for(i=0;i<=last;i++) {
    if(nodes[i]->nr_children==0) {
      br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      bc=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
    } else {
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];
      br=child1->u_rank+child2->u_rank;
      bc=child1->v_rank+child2->v_rank;
      br1=child1->u_rows+child1->u_rank;
      br2=child2->u_rows+child2->u_rank;
    }

    if(i!=dim_tree-1 || doRoot || saveVhat) {
      /* We expand V into its explicit form V=P*[I;E] */
      Ve=new T[bc*nodes[i]->v_rank]();
      for(j=0;j<nodes[i]->v_rank;j++)
        Ve[j*bc+j]=ONE;
      lacpy('N',nodes[i]->v_rows,nodes[i]->v_rank,nodes[i]->V_E,nodes[i]->v_rows,&Ve[nodes[i]->v_rank],bc);
      for(k=bc-1;k>=0;k--)
        if(k!=nodes[i]->V_perm[k]-1)
          swap(nodes[i]->v_rank,&Ve[k],bc,&Ve[nodes[i]->V_perm[k]-1],bc);
    }

    if(nodes[i]->nr_children!=0) {
      /* D = [ Dt_1 B12*Vt_2(rows2+1:end)' ; B21*Vt_1(rows1+1:end)' Dt_2] */
      nodes[i]->D=new T[br*br];

      /* (1,1) block */
      lacpy('N',child1->u_rank,child1->u_rank,child1->Dt,child1->u_rank,nodes[i]->D,br);

      /* (1,2) block */
      gemm('N','C',child1->u_rank,child2->u_rank,child2->v_rank,ONE,nodes[i]->B12,child1->u_rank,&child2->Vt[child2->u_rows],br2,ZERO,&nodes[i]->D[br*child1->u_rank],br);

      /* (2,1) block */
      gemm('N','C',child2->u_rank,child1->u_rank,child1->v_rank,ONE,nodes[i]->B21,child2->u_rank,&child1->Vt[child1->u_rows],br1,ZERO,&nodes[i]->D[child1->u_rank],br);

      /* (2,2) block */
      lacpy('N',child2->u_rank,child2->u_rank,child2->Dt,child2->u_rank,&nodes[i]->D[br*child1->u_rank+child1->u_rank],br);

      if(i!=dim_tree-1 || doRoot || saveVhat) {
        /* V = [Vt_1(rows1+1:end) 0 ; 0 Vt_2(rows2+1:end) ] * V (in array Ve) */

        /* Product into (new) V */
        tmp=new T[br*nodes[i]->v_rank];
        gemm('N','N',child1->u_rank,nodes[i]->v_rank,child1->v_rank,ONE,&child1->Vt[child1->u_rows],br1,Ve,bc,ZERO,tmp,br);
        gemm('N','N',child2->u_rank,nodes[i]->v_rank,child2->v_rank,ONE,&child2->Vt[child2->u_rows],br2,&Ve[child1->v_rank],bc,ZERO,&tmp[child1->u_rank],br);
        delete[] Ve;
        Ve=tmp;
      }
    }
    if((partial && i==last) || (i==dim_tree-1 && saveVhat)) {
      /* Vhat (if k is an interior node) of V (if k is a leaf)
       * stored in Vt and used for Schur complement computation.
       * Note that the following step is the LU factorization
       * and Ve is not destroyed, so we can safely point to it.
       */
      nodes[i]->Vt=Ve;
    }

    if(i==last && !doRoot) {
      /* Dense LU at root node */
      nodes[i]->L=new T[br*br];
      lacpy('N',br,br,nodes[i]->D,br,nodes[i]->L,br);
      nodes[i]->piv=new int[br];
      info=getrf(br,br,nodes[i]->L,br,nodes[i]->piv);
      if(info!=0)
        std::cout << "Something went wrong with LU at root node" << std::endl;
    } else {
      /* W = [-U_E I ; I 0 ] * U_perm' * D */
      /* U_perm' * D is stored in Dt temporarily */
      nodes[i]->Dt=new T[br*br];
      lacpy('N',br,br,nodes[i]->D,br,nodes[i]->Dt,br);
      for(k=0;k<br;k++)
        if(k!=nodes[i]->U_perm[k]-1)
          swap(br,&nodes[i]->Dt[k],br,&nodes[i]->Dt[nodes[i]->U_perm[k]-1],br);

      /* W = [ W1 ; W2 ] with W1 = Dt(u_rank+1:end,:)- U_E * Dt(1:u_rank,:) and W2 = Dt(1:u_rank,:) */
      nodes[i]->W=new T[br*br];
      if(nodes[i]->u_rows) {
        lacpy('N',nodes[i]->u_rows,br,&nodes[i]->Dt[nodes[i]->u_rank],br,nodes[i]->W,br);
        gemm('N','N',nodes[i]->u_rows,br,nodes[i]->u_rank,NONE,nodes[i]->U_E,nodes[i]->u_rows,nodes[i]->Dt,br,ONE,nodes[i]->W,br);
      }
      lacpy('N',nodes[i]->u_rank,br,nodes[i]->Dt,br,&nodes[i]->W[nodes[i]->u_rows],br);
      delete[] nodes[i]->Dt; nodes[i]->Dt=NULL;

      if(nodes[i]->u_rows) {
        /* LQ(W1): W1 = [L 0] * Q */
        nodes[i]->Q=new T[br*br]();
        lacpy('N',nodes[i]->u_rows,br,nodes[i]->W,br,nodes[i]->Q,br);
        tau=new T[br];
        gelqf(nodes[i]->u_rows,br,nodes[i]->Q,br,tau);

        /* Copy the L factor into nodes[i]->L */
        nodes[i]->L=new T[nodes[i]->u_rows*nodes[i]->u_rows]();
        lacpy('L',nodes[i]->u_rows,nodes[i]->u_rows,nodes[i]->Q,br,nodes[i]->L,nodes[i]->u_rows);

        /* Form the explicit Q factor from the Householder reflectors */
        xxglq(br,br,nodes[i]->u_rows,nodes[i]->Q,br,tau);
        delete[] tau;

        /* Vt = Q V ; V (explicit Ve) is not used afterwards */
        nodes[i]->Vt=new T[br*nodes[i]->v_rank];
        gemm('N','N',br,nodes[i]->v_rank,br,ONE,nodes[i]->Q,br,Ve,br,ZERO,nodes[i]->Vt,br);
        delete[] Ve;

        /* Dt = W(u_rows+1:end,:) * Q(u_rows+1:end,:)' */
        nodes[i]->Dt=new T[nodes[i]->u_rank*nodes[i]->u_rank];
        gemm('N','C',nodes[i]->u_rank,nodes[i]->u_rank,br,ONE,&nodes[i]->W[nodes[i]->u_rows],br,&nodes[i]->Q[nodes[i]->u_rows],br,ZERO,nodes[i]->Dt,nodes[i]->u_rank);
      } else {
        /* If u_rows=0, Vt=V, D=W, Q=I (not stored) */
        nodes[i]->Vt=new T[nodes[i]->u_rank*nodes[i]->v_rank];
        lacpy('N',nodes[i]->u_rank,nodes[i]->v_rank,Ve,nodes[i]->u_rank,nodes[i]->Vt,nodes[i]->u_rank);
        nodes[i]->Dt=new T[nodes[i]->u_rank*nodes[i]->u_rank];
        lacpy('N',nodes[i]->u_rank,nodes[i]->u_rank,nodes[i]->W,nodes[i]->u_rank,nodes[i]->Dt,nodes[i]->u_rank);

      }
    }

#if defined(debug)
    if(i!=dim_tree-1 || doRoot) {
      printMatrix("W",nodes[i]->W,br,br,br);
      if(nodes[i]->u_rows==0)
        std::cout << std::endl << "Q=I" << std::endl << std::endl;
      else {
        printMatrix("Q",nodes[i]->Q,br,br,br);
        printMatrix("L",nodes[i]->L,nodes[i]->u_rows,nodes[i]->u_rows,nodes[i]->u_rows);
      }
      printMatrix("Vt",nodes[i]->Vt,br,nodes[i]->v_rank,br);
    } else
      printMatrix("L",nodes[i]->L,br,br,br);
#endif

  }

  /* Clean-up: D for non-leaves and Dt are not needed anymore */
  if(!doRoot) {
    for(i=0;i<=last;i++) {
      if(nodes[i]->nr_children!=0) {
        delete[] nodes[i]->D; nodes[i]->D=NULL;
      }
      delete[] nodes[i]->Dt; nodes[i]->Dt=NULL;
    }
  }
}

#endif // HSS_SEQ_ULV_HPP
