/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#include <stdio.h>
#include "mex.h"
#include "StrumpackDensePackage.hpp"

const char *fieldNames[] = {"use_HSS",
                            "tol_HSS",
                            "min_rand_HSS", "max_rand_HSS", "lim_rand_HSS", "inc_rand_HSS",
                            "block_HSS", "levels_HSS",
                            "split_HSS",
                            "tol_IR", "steps_IR", "fast_IR",
                            "verbose",
                            "access_is_global",
                            "sched_strat",
                            "obj",
                            "ctxt",
                            "myid",
                            "np",
                            "nprow",
                            "npcol"
                           };
const int fieldNum = 21;

/* Block size for the 2D block-cyclic distribution */
const int nb2Dmex = 16;

template<typename T, typename S> void copystructfields(const mxArray *s, StrumpackDensePackage<T,S> *sdp) {

  mxArray *field;

  field=mxGetField(s,0,"use_HSS");
  sdp->use_HSS=mxGetPr(field)[0];

  field=mxGetField(s,0,"tol_HSS");
  sdp->tol_HSS=mxGetPr(field)[0];

  field=mxGetField(s,0,"min_rand_HSS");
  sdp->min_rand_HSS=mxGetPr(field)[0];

  field=mxGetField(s,0,"max_rand_HSS");
  sdp->max_rand_HSS=mxGetPr(field)[0];

  field=mxGetField(s,0,"lim_rand_HSS");
  sdp->lim_rand_HSS=mxGetPr(field)[0];

  field=mxGetField(s,0,"inc_rand_HSS");
  sdp->inc_rand_HSS=mxGetPr(field)[0];

  field=mxGetField(s,0,"block_HSS");
  sdp->block_HSS=mxGetPr(field)[0];

  field=mxGetField(s,0,"levels_HSS");
  sdp->levels_HSS=mxGetPr(field)[0];

  field=mxGetField(s,0,"split_HSS");
  sdp->split_HSS=mxGetPr(field)[0];

  field=mxGetField(s,0,"tol_IR");
  sdp->tol_IR=mxGetPr(field)[0];

  field=mxGetField(s,0,"steps_IR");
  sdp->steps_IR=mxGetPr(field)[0];

  field=mxGetField(s,0,"fast_IR");
  sdp->fast_IR=mxGetPr(field)[0];

  field=mxGetField(s,0,"verbose");
  sdp->verbose=mxGetPr(field)[0];

  field=mxGetField(s,0,"access_is_global");
  sdp->access_is_global=mxGetPr(field)[0];

  field=mxGetField(s,0,"sched_strat");
  sdp->sched_strat=mxGetPr(field)[0];

}

template<typename T> inline bool SDP_mex_is_complex();
template<> inline bool SDP_mex_is_complex<dcomplex>(){
 return true;
}
template<> inline bool SDP_mex_is_complex<double>(){
 return false;
}

template<typename T> void FromCentralizedTo2D(T *Acent, int m, int n, T**A, int ctxt, int *descA) {
  /* Redistribute a centralized m x n matrix into 2D block-cylic form.
   * If Acent is NULL, simply fill out the descriptor.
   * "ctxt" is an input context to be used for A.
   */
  int ctxtcent, ctxtglob;
  int myid, np;
  int nprow, npcol;
  int myrow, mycol;
  int nb;
  int idummy, ierr;
  int locr, locc;
  int descAcent[BLACSCTXTSIZE];

  MPI_Comm_rank(MPI_COMM_WORLD,&myid);
  MPI_Comm_size(MPI_COMM_WORLD,&np);
  nb=nb2Dmex;

  /* We initialize a context with only id 0 */
  blacs_get_(&IZERO,&IZERO,&ctxtcent);
  blacs_gridinit_(&ctxtcent,"R",&IONE,&IONE);

  /* We initialize a context with all the processes */
  blacs_get_(&IZERO,&IZERO,&ctxtglob);
  blacs_gridinit_(&ctxtglob,"R",&IONE,&np);

  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

  /* Distribute A into 2D block-cyclic form */
  if(myid==0) {
    descinit_(descAcent,&m,&n,&nb,&nb,&IZERO,&IZERO,&ctxtcent,&m,&ierr);
  } else {
    descset_(descAcent,&m,&n,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  if(myid<nprow*npcol) {
    locr=numroc_(&m,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    idummy=std::max(1,locr);
    *A=new T[locr*locc];
    descinit_(descA,&m,&n,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
  } else {
    *A=NULL;
    descset_(descA,&m,&n,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  if(Acent!=NULL)
    pgemr2d(m,n,Acent,IONE,IONE,descAcent,*A,IONE,IONE,descA,ctxtglob);

  if(!myid)
    blacs_gridexit_(&ctxtcent);
  blacs_gridexit_(&ctxtglob);

}

template<typename T> void From2DToReplicated(T *Adist, int m, int n, int ctxtdist, T**A) {
  /* Replicates a matrix distributed in 2D block-cylic form onto all the processes */
  int ctxt, ctxtglob;
  int myid, np;
  int nprow, npcol;
  int myrow, mycol;
  int nb;
  int idummy, ierr;
  int locr;
  int descAdist[BLACSCTXTSIZE], descA[BLACSCTXTSIZE];

  MPI_Comm_rank(MPI_COMM_WORLD,&myid);
  MPI_Comm_size(MPI_COMM_WORLD,&np);
  nb=nb2Dmex;

  /* We initialize a context with only id 0 */
  blacs_get_(&IZERO,&IZERO,&ctxt);
  blacs_gridinit_(&ctxt,"R",&IONE,&IONE);

  /* We initialize a context with all the processes */
  blacs_get_(&IZERO,&IZERO,&ctxtglob);
  blacs_gridinit_(&ctxtglob,"R",&IONE,&np);

  blacs_gridinfo_(&ctxtdist,&nprow,&npcol,&myrow,&mycol);

  /* Gathers A onto id 0 */
  if(myid==0) {
    *A=new T[m*n];
    descinit_(descA,&m,&n,&nb,&nb,&IZERO,&IZERO,&ctxt,&m,&ierr);
  } else {
    *A=NULL;
    descset_(descA,&m,&n,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  if(myid<nprow*npcol) {
    locr=numroc_(&m,&nb,&myrow,&IZERO,&nprow);
    idummy=std::max(1,locr);
    descinit_(descAdist,&m,&n,&nb,&nb,&IZERO,&IZERO,&ctxtdist,&idummy,&ierr);
  } else {
    descset_(descAdist,&m,&n,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  pgemr2d(m,n,Adist,IONE,IONE,descAdist,*A,IONE,IONE,descA,ctxtglob);

  /* Broadcast A */
  if(myid)
    *A=new T[m*n];
  MPI_Bcast((void*)*A,m*n,SDP_mex_is_complex<T>()?MPI_C_DOUBLE_COMPLEX:MPI_DOUBLE,IZERO,MPI_COMM_WORLD);

  if(!myid)
    blacs_gridexit_(&ctxt);
  blacs_gridexit_(&ctxtglob);

}
