/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef PAR_HQRINTERPOL_HPP
#define PAR_HQRINTERPOL_HPP

template<typename T, typename S> void par_HQRInterpol(int m, int n, T *A, int* descA, S tol, int *r, T **X, int **J, int **P, gridinfo *ginfo) {
  /* Parallel Interpolative Decomposition of A.
   * Decompose the m x n matrix A into A=A(:,J) [I X]*P^T
   * with J a set of columns of size r, and X rk x (n-rk).
   * Step 1: RRQR(A) => A P = Q R = Q [R1 R2]
   * Step 2: TRSM: A P = (Q R1) (R1^-1 R) => A = A(:,J)*[I X]*P^T
   *
   * Matrices are distributed and assumed to be stored in column-major format.
   * Matrix A is NOT modified.
   *
   * Matrix Q is initially a copy of A, then it contains the Q factor and is eventually discarded.
   * Matrix X has initial size n x n but is eventually reshaped into r x (n-r) (0-rows and identity columns are discarded).
   *
   * This is a Householder algorithm. It is a essentially a modified version of ScaLAPACK's
   * pxgeqpf with a threshold.

   */

  int ctxt, nprow, npcol, myrow, mycol, mb, nb;
  int i;
  int dummy, dummy2, ierr;
  int locrA, loccA, locrR, loccR;
  int descQ[BLACSCTXTSIZE], descR[BLACSCTXTSIZE];
  T *Q, *R;

  T ZERO=0;
  T  ONE=1;

#if defined(checkID)
  T *B, *C;
  S *work;
  S err;
#endif

#if defined(checkID)||defined(debug)
  int myid;
  MPI_Comm_rank(MPI_COMM_WORLD,&myid);
#endif

  ctxt=ginfo->ctxt;
  nprow=ginfo->nprow;
  npcol=ginfo->npcol;
  myrow=ginfo->myrow;
  mycol=ginfo->mycol;
  mb=ginfo->mb;
  nb=ginfo->nb;

  locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
  loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);

  /* Q: m x n, copy of A */
  Q=NULL;
  Q=new T[locrA*loccA];
  for(i=0;i<locrA*loccA;i++)
    Q[i]=A[i];
  dummy=std::max(1,locrA);
  descinit_(descQ,&m,&n,&mb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

  /* R: n x n */
  locrR=numroc_(&n,&mb,&myrow,&IZERO,&nprow);
  loccR=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
  R=NULL;
  R=new T[locrR*loccR];
  for(i=0;i<locrR*loccR;i++)
    R[i]=ZERO;
  dummy=std::max(1,locrR);
  descinit_(descR,&n,&n,&mb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

  /* J: indices of permuted colums */
  *J=new int[n];
  for(i=0;i<n;i++)
    (*J)[i]=i+1;

  /* P: permutation of columns */
  *P=new int[n];
  for(i=0;i<n;i++)
    (*P)[i]=i+1;

  /* Step 1: RRQR */
  pgeqpfmod(m,n,Q,IONE,IONE,descQ,*J,*P,r,tol,mb,nb,myrow,mycol,nprow,npcol);

  /* Copy the R factor into R */
  placpy('U',*r,n,Q,IONE,IONE,descQ,R,IONE,IONE,descR);

#if defined(debug)
  if(!myid) std::cout << "Rank = " << *r << std::endl << std::endl;
  printMatrix("Q: "+SSTR(myid),Q,locrA,loccA,locrA);
  printMatrix("R: "+SSTR(myid),R,locrR,loccR,locrR);
#endif

  delete[] Q;

  /* Step 2: TRSM and permutation: R1^-1 R = [I R1^-1 R2] = [I X] with R = [R1 R2], R1 r x r */
  ptrsm('L','U','N','N',*r,n-*r,ONE,R,IONE,IONE,descR,R,IONE,std::min(*r+1,n),descR);

  dummy=n-*r;
  locrR=numroc_(r,&mb,&myrow,&IZERO,&nprow);
  loccR=numroc_(&dummy,&nb,&mycol,&IZERO,&npcol);
  dummy2=std::max(1,locrR);
  *X=NULL;
  *X=new T[locrR*loccR];
  descinit_(descQ,r,&dummy,&mb,&nb,&IZERO,&IZERO,&ctxt,&dummy2,&ierr);
  pgeadd('N',*r,n-*r,ONE,R,IONE,*r+1,descR,ZERO,*X,IONE,IONE,descQ);
  delete[] R;

#if defined(debug)
  printMatrix("X: "+SSTR(myid),*X,locrR,loccR,locrR);
#endif

#if defined(checkID)
  /* Copy A into B. Compute C=A(:,J)*[I X] by copying A(:,J) into C(:,1:r) then
   * computing C(:,r+1:n)=C(:,1:r)*X. Then permute into C=C P^T.
   * Finally compute B=B-C=A-A(:,J)*[I X]*P^T and norm(B)/norm(A). */
  B=NULL;
  B=new T[locrA*loccA];
  for(i=0;i<locrA*loccA;i++)
    B[i]=A[i];

  C=NULL;
  C=new T[locrA*loccA];
  for(i=1;i<=*r;i++)
    pgeadd('N',m,1,ONE,A,IONE,(*J)[i-1],descA,ZERO,C,IONE,i,descA);
  pgemm('N','N',m,n-*r,*r,ONE,C,IONE,IONE,descA,*X,IONE,IONE,descQ,ZERO,C,IONE,*r+1,descA);
  for(i=n;i>=1;i--)
    if(i!=(*P)[i-1])
      pswap(m,C,IONE,(*P)[i-1],descA,IONE,C,IONE,i,descA,IONE);

  for(i=0;i<locrA*loccA;i++)
    B[i]=B[i]-C[i];

  work=NULL;
  work=new S[locrA];
  err=plange('I',m,n,B,IONE,IONE,descA,work)/plange('I',m,n,A,IONE,IONE,descA,work);
  if(!myid) std::cout << "Parallel RRQR: ||A-A(:,J)X||_inf/||A||_inf=" << err << std::endl << std::endl;
  delete[] B;
  delete[] C;
  delete[] work;
#endif

  int *Jtmp=new int[*r];
  for(int i=0;i<*r;i++)
    Jtmp[i]=(*J)[i];
  delete[] *J;
  *J=Jtmp;

  return;

}

#endif // PAR_HQRINTERPOL_HPP
